@@ -22,9 +22,10 @@ import warnings
 import numpy as np
 import pandas as pd
 
-from ..globalvar import SOURCE, TARGET, WEIGHT
+from ..globalvar import SOURCE, TARGET, WEIGHT, TYPE_ATTR_NAME
 from .element_data import NodeData, EdgeData
 from .validation import comma_sep, require_dataframe_has_columns
+from .utils import is_real_iterable
 
 
 class ColumnarConverter:
@@ -35,26 +36,42 @@ class ColumnarConverter:
     Args:
         name (str): the name of the argument for error messages
         default_type (hashable): the default type to use for data without a type
+        type_column (hashable, optional): the name of the type column, if one is being used
         column_defaults (dict of hashable to any): any default values for columns (using names before renaming!)
         selected_columns (dict of hashable to hashable): renamings for columns, mapping original name to new name
         allow_features (bool): if True, columns that aren't selected are returned as a numpy feature matrix
         dtype (str or numpy dtype): the data type to use for the feature matrices
+        transform_columns (dict of hashable to callable): column transformations, maps column name to transform
     """
 
     def __init__(
         self,
         name,
         default_type,
+        type_column,
         column_defaults,
         selected_columns,
         allow_features,
+        transform_columns,
         dtype=None,
     ):
+        if type_column is not None:
+            if allow_features:
+                raise ValueError(
+                    f"allow_features: expected no features when using a type column ({type_column!r}), found them to be allowed"
+                )
+            if type_column not in selected_columns:
+                raise ValueError(
+                    f"selected_columns: expected type column ({type_column!r}) to be included when using, found only {comma_sep(list(selected_columns.keys()))}"
+                )
+
         self._parent_name = name
+        self.type_column = type_column
         self.column_defaults = column_defaults
         self.selected_columns = selected_columns
         self.default_type = default_type
         self.allow_features = allow_features
+        self.transform_columns = transform_columns
         self.dtype = dtype
 
     def name(self, type_name=None):
@@ -65,7 +82,7 @@ class ColumnarConverter:
     def _convert_single(self, type_name, data):
         if not isinstance(data, pd.DataFrame):
             raise TypeError(
-                f"{self.name(type_name)}: expected pandas DataFrame, found {type(data)}"
+                f"{self.name(type_name)}: expected pandas DataFrame, found {type(data).__name__}"
             )
 
         existing = set(self.selected_columns).intersection(data.columns)
@@ -87,8 +104,15 @@ class ColumnarConverter:
             self.name(type_name), known, self.selected_columns
         )
 
+        for column, transform in self.transform_columns.items():
+            known[column] = transform(known[column])
+
         if self.allow_features:
-            features = other.to_numpy(dtype=self.dtype)
+            # to_numpy returns an unspecified order but it's Fortran in practice. Row-level bulk
+            # operations are more common (e.g. slicing out a couple of row, when sampling a few
+            # nodes) than column-level ones so having rows be contiguous (C order) is much more
+            # efficient.
+            features = np.ascontiguousarray(other.to_numpy(dtype=self.dtype))
         elif len(other.columns) == 0:
             features = None
         else:
@@ -98,57 +122,157 @@ class ColumnarConverter:
 
         return known.rename(columns=self.selected_columns), features
 
+    def _shared_and_starts_from_singles(self, singles):
+        rows_so_far = 0
+        type_starts = []
+        type_dfs = []
+
+        for type_name in sorted(singles.keys()):
+            type_data, _ = singles[type_name]
+
+            type_starts.append((type_name, rows_so_far))
+            type_dfs.append(type_data)
+
+            rows_so_far += len(type_data)
+
+        if type_dfs:
+            shared = pd.concat(type_dfs)
+        else:
+            # there was no input types and thus no input elements, so create a dummy dataframe, that
+            # is maximally flexible by using a "minimal"/highly-promotable type
+            shared = pd.DataFrame(
+                columns=self.selected_columns.values(), dtype=np.uint8
+            )
+
+        return shared, type_starts
+
+    def _convert_with_type_column(self, data):
+        # we've got a type column, so there's no dictionaries or separate dataframes. We just need
+        # to make sure things are arranged right, i.e. nodes of each type are contiguous, and
+        # 'range(...)' objects describing each one.
+        known, features = self._convert_single(None, data)
+        assert features is None
+
+        # the column we see in `known` is after being selected/renamed
+        selected_type_column = self.selected_columns[self.type_column]
+
+        known.sort_values(selected_type_column, inplace=True)
+
+        # the shared data doesn't use the type column; that info is encoded in `type_ranges` below
+        shared = known.drop(columns=selected_type_column)
+
+        # deduce the type ranges based on the first index of each of the known values
+        types, first_occurance = np.unique(
+            known[selected_type_column], return_index=True
+        )
+
+        type_starts = [
+            (type_name, start) for type_name, start in zip(types, first_occurance)
+        ]
+
+        # per the assert above, the features are None for every type
+        features = {type_name: None for type_name in types}
+
+        return shared, type_starts, features
+
     def convert(self, elements):
+        if self.type_column is not None:
+            return self._convert_with_type_column(elements)
+
         if isinstance(elements, pd.DataFrame):
             elements = {self.default_type: elements}
 
         if not isinstance(elements, dict):
-            raise TypeError(f"{self.name()}: expected dict, found {type(elements)}")
+            raise TypeError(
+                f"{self.name()}: expected dict, found {type(elements).__name__}"
+            )
 
         singles = {
             type_name: self._convert_single(type_name, data)
             for type_name, data in elements.items()
         }
-        return (
-            {type_name: shared for type_name, (shared, _) in singles.items()},
-            {type_name: features for type_name, (_, features) in singles.items()},
-        )
+
+        shared, type_starts = self._shared_and_starts_from_singles(singles)
+        features = {type_name: features for type_name, (_, features) in singles.items()}
+        return (shared, type_starts, features)
 
 
 def convert_nodes(data, *, name, default_type, dtype) -> NodeData:
     converter = ColumnarConverter(
         name,
         default_type,
+        type_column=None,
         column_defaults={},
         selected_columns={},
         allow_features=True,
+        transform_columns={},
         dtype=dtype,
     )
-    nodes, node_features = converter.convert(data)
-    return NodeData(nodes, node_features)
+    nodes, type_starts, node_features = converter.convert(data)
+    return NodeData(nodes, type_starts, node_features)
 
 
 DEFAULT_WEIGHT = np.float32(1)
 
 
 def convert_edges(
-    data, *, name, default_type, source_column, target_column, weight_column,
+    data,
+    *,
+    name,
+    default_type,
+    source_column,
+    target_column,
+    weight_column,
+    type_column,
+    nodes,
 ):
+    def _node_ids_to_iloc(node_ids):
+        try:
+            return nodes.ids.to_iloc(node_ids, strict=True)
+        except KeyError as e:
+            missing_values = e.args[0]
+            if not is_real_iterable(missing_values):
+                missing_values = [missing_values]
+            missing_values = pd.unique(missing_values)
+
+            raise ValueError(
+                f"edges: expected all source and target node IDs to be contained in `nodes`, "
+                f"found some missing: {comma_sep(missing_values)}"
+            )
+
+    selected = {
+        source_column: SOURCE,
+        target_column: TARGET,
+        weight_column: WEIGHT,
+    }
+
+    if type_column is not None:
+        selected[type_column] = TYPE_ATTR_NAME
+
     converter = ColumnarConverter(
         name,
         default_type,
+        type_column=type_column,
         column_defaults={weight_column: DEFAULT_WEIGHT},
-        selected_columns={
-            source_column: SOURCE,
-            target_column: TARGET,
-            weight_column: WEIGHT,
+        selected_columns=selected,
+        transform_columns={
+            source_column: _node_ids_to_iloc,
+            target_column: _node_ids_to_iloc,
         },
         allow_features=False,
     )
-    edges, edge_features = converter.convert(data)
+    edges, type_starts, edge_features = converter.convert(data)
+
+    # validation:
     assert all(features is None for features in edge_features.values())
 
-    return EdgeData(edges)
+    weight_col = edges[WEIGHT]
+    if not pd.api.types.is_numeric_dtype(weight_col):
+        raise TypeError(
+            f"{converter.name()}: expected weight column {weight_column!r} to be numeric, found dtype '{weight_col.dtype}'"
+        )
+
+    return EdgeData(edges, type_starts)
 
 
 SingleTypeNodeIdsAndFeatures = namedtuple(
@@ -201,7 +325,7 @@ def _features_from_attributes(node_type, ids, values, dtype):
     return matrix
 
 
-def _features_from_node_data(nodes, data, dtype):
+def _features_from_node_data(nodes, node_type_default, data, dtype):
     if isinstance(data, dict):
 
         def single(node_type):
@@ -251,8 +375,10 @@ def _features_from_node_data(nodes, data, dtype):
                 "When there is more than one node type, pass node features as a dictionary."
             )
 
-        node_type = next(iter(nodes))
-        return _features_from_node_data(nodes, {node_type: data}, dtype)
+        node_type = next(iter(nodes), node_type_default)
+        return _features_from_node_data(
+            nodes, node_type_default, {node_type: data}, dtype
+        )
     elif isinstance(data, (Iterable, list)):
         id_to_data = dict(data)
         return {
@@ -305,7 +431,9 @@ def from_networkx(
             for node_type, node_info in nodes.items()
         }
     else:
-        node_frames = _features_from_node_data(nodes, node_features, dtype)
+        node_frames = _features_from_node_data(
+            nodes, node_type_default, node_features, dtype
+        )
 
     edges = nx.to_pandas_edgelist(graph, source=SOURCE, target=TARGET)
     _fill_or_assign(edges, edge_type_attr, edge_type_default)
