@@ -131,7 +131,7 @@ class StellarGraph:
     Notice the ``foo`` node has one feature ``x``, while the ``bar`` nodes have 2 features ``y`` and
     ``z``. A heterogeneous graph can have different features for each type.
 
-    Edges of different types work in the same way. example instance, if edges have different types based
+    Edges of different types can work in the same way. For instance, if edges have different types based
     on their orientation::
 
         horizontal_edges = pd.DataFrame(
@@ -151,12 +151,34 @@ class StellarGraph:
             {"h": horizontal_edges, "v": vertical_edges, "d": diagonal_edges}
         )
 
+    Alternatively, a single DataFrame can be provided, with an additional column of the type. This
+    column is specified by passing the ``edge_type_column`` argument::
+
+        orientation_edges = pd.DataFrame(
+            {
+                "source": ["a", "b", "c", "d", "a"],
+                "target": ["b", "c", "d", "a", "c"],
+                "type": ["h", "v", "h", "v", "d"]
+            }
+        )
+
+        StellarGraph(nodes, orientation_edges, edge_type_column="type")
+
     .. note::
 
         The IDs of nodes must be unique across all types: for example, it is an error to have a node
         0 of type ``a``, and a node 0 of type ``b``. IDs of edges must also be unique across all
         types.
 
+    .. _iloc-explanation:
+
+    This type stores the external IDs for nodes and edges as *ilocs* (internal or integer locations,
+    similar to Pandas), which are sequential integers that allow for efficient storage and indexing.
+    For convenience, methods here will traffic in the external ID values and transparently convert to
+    and from ilocs as required internally. Many of these methods also have a ``use_ilocs`` parameter
+    that allows for explicitly switching the methods to consume and return ilocs directly, cutting out
+    the conversion overhead.
+
     .. seealso:: :meth:`from_networkx` for construction from a NetworkX graph.
 
     Args:
@@ -193,6 +215,10 @@ class StellarGraph:
             The name of the column in each of the ``edges`` DataFrames to use as the weight of
             edges. If the column does not exist in any of them, it is defaulted to ``1``.
 
+        edge_type_column (str, optional):
+            The name of the column in the ``edges`` DataFrame to use as the edge type (if this is
+            set, ``edges`` must be a single DataFrame, not a dictionary).
+
         node_type_default (str, optional):
             The default node type to use, if ``nodes`` is passed as a DataFrame (not a ``dict``).
 
@@ -221,6 +247,7 @@ class StellarGraph:
         source_column=globalvar.SOURCE,
         target_column=globalvar.TARGET,
         edge_weight_column=globalvar.WEIGHT,
+        edge_type_column=None,
         node_type_default=globalvar.NODE_TYPE_DEFAULT,
         edge_type_default=globalvar.EDGE_TYPE_DEFAULT,
         dtype="float32",
@@ -267,9 +294,19 @@ class StellarGraph:
                 dtype=dtype,
             )
 
+        if nodes is None:
+            nodes_after_inference = self._infer_nodes_from_edges(
+                edges, source_column, target_column
+            )
+            nodes = pd.DataFrame([], index=nodes_after_inference)
+
         if edges is None:
             edges = {}
 
+        self._nodes = convert.convert_nodes(
+            nodes, name="nodes", default_type=node_type_default, dtype=dtype,
+        )
+
         self._is_directed = is_directed
         self._edges = convert.convert_edges(
             edges,
@@ -278,41 +315,34 @@ class StellarGraph:
             source_column=source_column,
             target_column=target_column,
             weight_column=edge_weight_column,
+            type_column=edge_type_column,
+            nodes=self._nodes,
         )
 
         nodes_from_edges = pd.unique(
             np.concatenate([self._edges.targets, self._edges.sources])
         )
 
-        if nodes is None:
-            nodes_after_inference = pd.DataFrame([], index=nodes_from_edges)
+    @staticmethod
+    def _infer_nodes_from_edges(edges, source_column, target_column):
+        # `convert_edges` nicely flags any errors in edges; inference here is lax rather than duplicate that
+        if isinstance(edges, dict):
+            dataframes = edges.values()
         else:
-            nodes_after_inference = nodes
+            dataframes = [edges]
 
-        self._nodes = convert.convert_nodes(
-            nodes_after_inference,
-            name="nodes",
-            default_type=node_type_default,
-            dtype=dtype,
-        )
+        found_columns = [
+            type_edges[column]
+            for type_edges in dataframes
+            if isinstance(type_edges, pd.DataFrame)
+            for column in [source_column, target_column]
+            if column in type_edges.columns
+        ]
 
-        if nodes is not None:
-            # check for dangling edges: make sure the explicitly-specified nodes parameter includes every
-            # node mentioned in the edges
-            try:
-                self._nodes.ids.to_iloc(
-                    nodes_from_edges, smaller_type=False, strict=True,
-                )
-            except KeyError as e:
-                missing_values = e.args[0]
-                if not is_real_iterable(missing_values):
-                    missing_values = [missing_values]
-                missing_values = pd.unique(missing_values)
+        if found_columns:
+            return pd.unique(np.concatenate(found_columns))
 
-                raise ValueError(
-                    f"edges: expected all source and target node IDs to be contained in `nodes`, "
-                    f"found some missing: {comma_sep(missing_values)}"
-                )
+        return []
 
     @staticmethod
     def from_networkx(
@@ -474,50 +504,61 @@ class StellarGraph:
         """
         return len(self._edges)
 
-    def nodes(self, node_type=None) -> Iterable[Any]:
+    def nodes(self, node_type=None, use_ilocs=False) -> Iterable[Any]:
         """
         Obtains the collection of nodes in the graph.
 
         Args:
             node_type (hashable, optional): a type of nodes that exist in the graph
+            use_ilocs (bool): if True return :ref:`node ilocs <iloc-explanation>` as a ``range`` object
 
         Returns:
             All the nodes in the graph if ``node_type`` is ``None``, otherwise all the nodes in the
             graph of type ``node_type``.
         """
         if node_type is None:
-            return self._nodes.ids.pandas_index
+            all_ids = self._nodes.ids.pandas_index
+            if use_ilocs:
+                return range(self.number_of_nodes())
+            return all_ids
 
         ilocs = self._nodes.type_range(node_type)
+        if use_ilocs:
+            return ilocs
         return self._nodes.ids.from_iloc(ilocs)
 
     def edges(
-        self, include_edge_type=False, include_edge_weight=False
+        self, include_edge_type=False, include_edge_weight=False, use_ilocs=False
     ) -> Iterable[Any]:
         """
         Obtains the collection of edges in the graph.
 
         Args:
-            include_edge_type (bool): A flag that indicates whether to return edge types
-            of format (node 1, node 2, edge type) or edge pairs of format (node 1, node 2).
-            include_edge_weight (bool): A flag that indicates whether to return edge weights.
-            Weights are returned in a separate list.
+            include_edge_type (bool):
+                A flag that indicates whether to return edge types of format (node 1, node 2, edge
+                type) or edge pairs of format (node 1, node 2).
+            include_edge_weight (bool):
+                A flag that indicates whether to return edge weights.  Weights are returned in a
+                separate list.
+            use_ilocs (bool): if True return :ref:`node ilocs <iloc-explanation>`
 
         Returns:
             The graph edges. If edge weights are included then a tuple of (edges, weights)
         """
         # FIXME: these would be better returned as the 2 or 3 arrays directly, rather than tuple-ing
         # (the same applies to all other instances of zip in this file)
+        sources = self._edges.sources
+        targets = self._edges.targets
+
+        if not use_ilocs:
+            sources = self._nodes.ids.from_iloc(sources)
+            targets = self._nodes.ids.from_iloc(targets)
+
         if include_edge_type:
-            edges = list(
-                zip(
-                    self._edges.sources,
-                    self._edges.targets,
-                    self._edges.type_of_iloc(slice(None)),
-                )
-            )
+            types = self._edges.type_of_iloc(slice(None))
+            edges = list(zip(sources, targets, types))
         else:
-            edges = list(zip(self._edges.sources, self._edges.targets))
+            edges = list(zip(sources, targets))
 
         if include_edge_weight:
             return edges, self._edges.weights
@@ -538,19 +579,22 @@ class StellarGraph:
         return node in self._nodes
 
     def _transform_edges(
-        self, other_node_id, ilocs, include_edge_weight, filter_edge_types
+        self, other_node, ilocs, include_edge_weight, filter_edge_types, use_ilocs
     ):
         if include_edge_weight:
             weights = self._edges.weights[ilocs]
         else:
             weights = None
 
+        if not use_ilocs:
+            other_node = self._nodes.ids.from_iloc(other_node)
+
         if filter_edge_types is not None:
             filter_edge_type_ilocs = self._edges.types.to_iloc(filter_edge_types)
             edge_type_ilocs = self._edges.type_ilocs[ilocs]
             correct_type = np.isin(edge_type_ilocs, filter_edge_type_ilocs)
 
-            other_node_id = other_node_id[correct_type]
+            other_node = other_node[correct_type]
             if weights is not None:
                 weights = weights[correct_type]
 
@@ -558,13 +602,13 @@ class StellarGraph:
         if weights is not None:
             return [
                 NeighbourWithWeight(node, weight)
-                for node, weight in zip(other_node_id, weights)
+                for node, weight in zip(other_node, weights)
             ]
 
-        return list(other_node_id)
+        return list(other_node)
 
     def neighbors(
-        self, node: Any, include_edge_weight=False, edge_types=None
+        self, node: Any, include_edge_weight=False, edge_types=None, use_ilocs=False
     ) -> Iterable[Any]:
         """
         Obtains the collection of neighbouring nodes connected
@@ -576,20 +620,26 @@ class StellarGraph:
                 output is a named tuple with fields `node` (the node ID) and `weight` (the edge weight)
             edge_types (list of hashable, optional): If provided, only traverse the graph
                 via the provided edge types when collecting neighbours.
+            use_ilocs (bool): if True `node` is treated as a :ref:`node iloc <iloc-explanation>` and the ilocs of
+                each neighbour is returned.
 
         Returns:
             iterable: The neighbouring nodes.
         """
-        ilocs = self._edges.edge_ilocs(node, ins=True, outs=True)
-        source = self._edges.sources[ilocs]
-        target = self._edges.targets[ilocs]
-        other_node_id = np.where(source == node, target, source)
+        if not use_ilocs:
+            node = self._nodes.ids.to_iloc([node])[0]
+
+        edge_ilocs = self._edges.edge_ilocs(node, ins=True, outs=True)
+        source = self._edges.sources[edge_ilocs]
+        target = self._edges.targets[edge_ilocs]
+        other_node = np.where(source == node, target, source)
+
         return self._transform_edges(
-            other_node_id, ilocs, include_edge_weight, edge_types
+            other_node, edge_ilocs, include_edge_weight, edge_types, use_ilocs
         )
 
     def in_nodes(
-        self, node: Any, include_edge_weight=False, edge_types=None
+        self, node: Any, include_edge_weight=False, edge_types=None, use_ilocs=False
     ) -> Iterable[Any]:
         """
         Obtains the collection of neighbouring nodes with edges
@@ -602,6 +652,8 @@ class StellarGraph:
                 output is a named tuple with fields `node` (the node ID) and `weight` (the edge weight)
             edge_types (list of hashable, optional): If provided, only traverse the graph
                 via the provided edge types when collecting neighbours.
+            use_ilocs (bool): if True `node` is treated as a :ref:`node iloc <iloc-explanation>` and the ilocs of each neighbour is
+                returned.
 
         Returns:
             iterable: The neighbouring in-nodes.
@@ -609,15 +661,23 @@ class StellarGraph:
         if not self.is_directed():
             # all edges are both incoming and outgoing for undirected graphs
             return self.neighbors(
-                node, include_edge_weight=include_edge_weight, edge_types=edge_types
+                node,
+                include_edge_weight=include_edge_weight,
+                edge_types=edge_types,
+                use_ilocs=use_ilocs,
             )
 
-        ilocs = self._edges.edge_ilocs(node, ins=True, outs=False)
-        source = self._edges.sources[ilocs]
-        return self._transform_edges(source, ilocs, include_edge_weight, edge_types)
+        if not use_ilocs:
+            node = self._nodes.ids.to_iloc([node])[0]
+        edge_ilocs = self._edges.edge_ilocs(node, ins=True, outs=False)
+        source = self._edges.sources[edge_ilocs]
+
+        return self._transform_edges(
+            source, edge_ilocs, include_edge_weight, edge_types, use_ilocs
+        )
 
     def out_nodes(
-        self, node: Any, include_edge_weight=False, edge_types=None
+        self, node: Any, include_edge_weight=False, edge_types=None, use_ilocs=False
     ) -> Iterable[Any]:
         """
         Obtains the collection of neighbouring nodes with edges
@@ -630,6 +690,8 @@ class StellarGraph:
                 output is a named tuple with fields `node` (the node ID) and `weight` (the edge weight)
             edge_types (list of hashable, optional): If provided, only traverse the graph
                 via the provided edge types when collecting neighbours.
+            use_ilocs (bool): if True `node` is treated as a :ref:`node iloc <iloc-explanation>` and the ilocs of each neighbour is
+                returned.
 
         Returns:
             iterable: The neighbouring out-nodes.
@@ -637,12 +699,21 @@ class StellarGraph:
         if not self.is_directed():
             # all edges are both incoming and outgoing for undirected graphs
             return self.neighbors(
-                node, include_edge_weight=include_edge_weight, edge_types=edge_types
+                node,
+                include_edge_weight=include_edge_weight,
+                edge_types=edge_types,
+                use_ilocs=use_ilocs,
             )
 
-        ilocs = self._edges.edge_ilocs(node, ins=False, outs=True)
-        target = self._edges.targets[ilocs]
-        return self._transform_edges(target, ilocs, include_edge_weight, edge_types)
+        if not use_ilocs:
+            node = self._nodes.ids.to_iloc([node])[0]
+
+        edge_ilocs = self._edges.edge_ilocs(node, ins=False, outs=True)
+        target = self._edges.targets[edge_ilocs]
+
+        return self._transform_edges(
+            target, edge_ilocs, include_edge_weight, edge_types, use_ilocs
+        )
 
     def nodes_of_type(self, node_type=None):
         """
@@ -662,7 +733,7 @@ class StellarGraph:
         )
         return list(self.nodes(node_type=node_type))
 
-    def node_type(self, node):
+    def node_type(self, node, use_ilocs=False):
         """
         Get the type of the node
 
@@ -673,8 +744,9 @@ class StellarGraph:
             Node type
         """
         nodes = [node]
-        node_ilocs = self._nodes.ids.to_iloc(nodes, strict=True)
-        type_sequence = self._nodes.type_of_iloc(node_ilocs)
+        if not use_ilocs:
+            nodes = self._nodes.ids.to_iloc(nodes, strict=True)
+        type_sequence = self._nodes.type_of_iloc(nodes)
 
         assert len(type_sequence) == 1
         return type_sequence[0]
@@ -689,6 +761,39 @@ class StellarGraph:
         """
         return set(self._nodes.types.pandas_index)
 
+    def unique_node_type(self, error_message=None):
+        """
+        Return the unique node type, for a homogeneous-node graph.
+
+        Args:
+            error_message (str, optional): a custom message to use for the exception; this can use
+                the ``%(found)s`` placeholder to insert the real sequence of node types.
+
+        Returns:
+            If this graph has only one node type, this returns that node type, otherwise it raises a
+            ``ValueError`` exception.
+        """
+
+        all_types = self._nodes.types.pandas_index
+        if len(all_types) == 1:
+            return all_types[0]
+
+        found = comma_sep(all_types)
+        if error_message is None:
+            error_message = (
+                "Expected only one node type for 'unique_node_type', found: %(found)s"
+            )
+
+        raise ValueError(error_message % {"found": found})
+
+    @property
+    def edge_types(self):
+        """
+        Returns:
+            a sequence of all edge types in the graph
+        """
+        return self._edges.types.pandas_index
+
     def node_feature_sizes(self, node_types=None):
         """
         Get the feature sizes for the specified node types.
@@ -722,39 +827,103 @@ class StellarGraph:
 
         # TODO: check the feature node_ids against the graph node ids?
 
-    def node_features(self, nodes, node_type=None):
+    def node_ids_to_ilocs(self, nodes):
         """
-        Get the numeric feature vectors for the specified node or nodes.
-        If the node type is not specified the node types will be found
-        for all nodes. It is therefore important to supply the ``node_type``
-        for this method to be fast.
+        Get the :ref:`node ilocs <iloc-explanation>` for the specified node or nodes.
 
         Args:
-            nodes (list or hashable): Node ID or list of node IDs
-            node_type (hashable): the type of the nodes.
+            nodes (list or hashable): node IDs
 
         Returns:
-            Numpy array containing the node features for the requested nodes.
+            Numpy array containing the indices for the requested nodes.
         """
+        return self._nodes.ids.to_iloc(nodes, strict=True)
+
+    def node_ilocs_to_ids(self, node_ilocs):
+        """
+        Get the node ids for the specified :ref:`node ilocs <iloc-explanation>`.
+
+        Args:
+            node_ilocs (list or hashable): :ref:`node ilocs <iloc-explanation>`
+
+        Returns:
+            Numpy array containing the node ids for the requested nodes.
+        """
+        return self._nodes.ids.from_iloc(node_ilocs)
+
+    def node_features(self, nodes=None, node_type=None, use_ilocs=False):
+        """
+        Get the numeric feature vectors for the specified nodes or node type.
+
+        For graphs with a single node type:
+
+        - ``graph.node_features()`` to retrieve features of all nodes, in the same order as
+          ``graph.nodes()``.
+
+        - ``graph.node_features(nodes=some_node_ids)`` to retrieve features for each node in
+          ``some_node_ids``.
+
+        For graphs with multiple node types:
+
+        - ``graph.node_features(node_type=some_type)`` to retrieve features of all nodes of type
+          ``some_type``, in the same order as ``graph.nodes(node_type=some_type)``.
+
+        - ``graph.node_features(nodes=some_node_ids, node_type=some_type)`` to retrieve features for
+          each node in ``some_node_ids``. All of the chosen nodes must be of type ``some_type``.
+
+        - ``graph.node_features(nodes=some_node_ids)`` to retrieve features for each node in
+          ``some_node_ids``. All of the chosen nodes must be of the same type, which will be
+          inferred. This will be slower than providing the node type explicitly in the previous example.
+
+        Args:
+            nodes (list or hashable, optional): Node ID or list of node IDs, all of the same type
+            node_type (hashable, optional): the type of the nodes.
+
+        Returns:
+            Numpy array containing the node features for the requested nodes or node type.
+        """
+        if nodes is None:
+            if node_type is None:
+                node_type = self.unique_node_type(
+                    "node_type: in a non-homogeneous graph, expected a node type and/or 'nodes' to be passed; found neither 'node_type' nor 'nodes', and the graph has node types: %(found)s"
+                )
+
+            return self._nodes.features_of_type(node_type)
+
         nodes = np.asarray(nodes)
 
-        node_ilocs = self._nodes.ids.to_iloc(nodes)
+        if len(nodes) == 0:
+            # empty lists are cast to a default array type of float64 -
+            # must manually specify integer type if empty, in which case we can pretend we received ilocs
+            node_ilocs = np.asarray(nodes, dtype=np.uint8)
+            use_ilocs = True
+        elif use_ilocs:
+            nodes = np.asarray(nodes)
+            node_ilocs = np.asarray(nodes)
+        else:
+            nodes = np.asarray(nodes)
+            node_ilocs = self._nodes.ids.to_iloc(np.asarray(nodes))
+
         valid = self._nodes.ids.is_valid(node_ilocs)
         all_valid = valid.all()
         valid_ilocs = node_ilocs if all_valid else node_ilocs[valid]
 
         if node_type is None:
-            # infer the type based on the valid nodes
-            types = np.unique(self._nodes.type_of_iloc(valid_ilocs))
+            try:
+                # no inference required in a homogeneous-node graph
+                node_type = self.unique_node_type()
+            except ValueError:
+                # infer the type based on the valid nodes
+                types = np.unique(self._nodes.type_of_iloc(valid_ilocs))
 
-            if len(types) == 0:
-                raise ValueError(
-                    "must have at least one node for inference, if `node_type` is not specified"
-                )
-            if len(types) > 1:
-                raise ValueError("all nodes must have the same type")
+                if len(types) == 0:
+                    raise ValueError(
+                        "must have at least one node for inference, if `node_type` is not specified"
+                    )
+                if len(types) > 1:
+                    raise ValueError("all nodes must have the same type")
 
-            node_type = types[0]
+                node_type = types[0]
 
         if all_valid:
             return self._nodes.features(node_type, valid_ilocs)
@@ -767,8 +936,9 @@ class StellarGraph:
         # instead use an impossible integer (e.g. 2**64 - 1)
 
         # everything that's not the sentinel should be valid
-        non_nones = nodes != None
-        self._nodes.ids.require_valid(nodes[non_nones], node_ilocs[non_nones])
+        if not use_ilocs:
+            non_nones = nodes != None
+            self._nodes.ids.require_valid(nodes[non_nones], node_ilocs[non_nones])
 
         sampled = self._nodes.features(node_type, valid_ilocs)
         features = np.zeros((len(nodes), sampled.shape[1]))
@@ -780,12 +950,12 @@ class StellarGraph:
     # Computationally intensive methods:
 
     def _edge_type_iloc_triples(self, selector=slice(None), stacked=False):
-        source_ilocs = self._nodes.ids.to_iloc(self._edges.sources[selector])
+        source_ilocs = self._edges.sources[selector]
         source_type_ilocs = self._nodes.type_ilocs[source_ilocs]
 
         rel_type_ilocs = self._edges.type_ilocs[selector]
 
-        target_ilocs = self._nodes.ids.to_iloc(self._edges.targets[selector])
+        target_ilocs = self._edges.targets[selector]
         target_type_ilocs = self._nodes.type_ilocs[target_ilocs]
 
         all_ilocs = source_type_ilocs, rel_type_ilocs, target_type_ilocs
@@ -805,21 +975,16 @@ class StellarGraph:
             self._nodes.types.from_iloc(tgt_ilocs),
         )
 
-    def _unique_type_triples(self, *, return_counts, selector=slice(None)):
+    def _unique_type_triples(self, selector=slice(None)):
         all_type_ilocs = self._edge_type_iloc_triples(selector, stacked=True)
 
         if len(all_type_ilocs) == 0:
             # FIXME(https://github.com/numpy/numpy/issues/15559): if there's no edges, np.unique is
             # being called on a shape=(0, 3) ndarray, and hits "ValueError: cannot reshape array of
             # size 0 into shape (0,newaxis)", so we manually reproduce what would be returned
-            if return_counts:
-                ret = None, [], []
-            else:
-                ret = None, []
+            ret = None, []
         else:
-            ret = np.unique(
-                all_type_ilocs, axis=0, return_index=True, return_counts=return_counts
-            )
+            ret = np.unique(all_type_ilocs, axis=0, return_index=True)
 
         edge_ilocs = ret[1]
         # we've now got the indices for an edge with each triple, along with the counts of them, so
@@ -827,11 +992,20 @@ class StellarGraph:
         # getting the actual type for each type iloc in the triples)
         unique_ets = self._edge_type_triples(edge_ilocs)
 
-        if return_counts:
-            return zip(*unique_ets, ret[2])
-
         return zip(*unique_ets)
 
+    def _edge_metrics_by_type_triple(self, metrics):
+        src_ty, rel_ty, tgt_ty = self._edge_type_triples()
+        df = pd.DataFrame(
+            {
+                "src_ty": src_ty,
+                "rel_ty": rel_ty,
+                "tgt_ty": tgt_ty,
+                "weight": self._edges.weights,
+            }
+        )
+        return df.groupby(["src_ty", "rel_ty", "tgt_ty"]).agg(metrics)["weight"]
+
     def info(self, show_attributes=None, sample=None, truncate=20):
         """
         Return an information string summarizing information on the current graph.
@@ -895,6 +1069,16 @@ class StellarGraph:
                 edge_types = "none"
             return f"{nt}: [{count}]\n    Features: {feature_text}\n    Edge types: {edge_types}"
 
+        def edge_type_info(et, metrics):
+            if metrics.min == metrics.max:
+                weights_text = (
+                    "all 1 (default)" if metrics.min == 1 else f"all {metrics.min:.6g}"
+                )
+            else:
+                weights_text = f"range=[{metrics.min:.6g}, {metrics.max:.6g}], mean={metrics.mean:.6g}, std={metrics.std:.6g}"
+
+            return f"{str_edge_type(et)}: [{metrics.count}]\n        Weights: {weights_text}"
+
         # sort the node types in decreasing order of frequency
         node_types = sorted(
             ((len(self.nodes(node_type=nt)), nt) for nt in gs.node_types), reverse=True
@@ -906,18 +1090,19 @@ class StellarGraph:
             sep="\n  ",
         )
 
-        # FIXME: it would be better for the schema to just include the counts directly
-        unique_ets = self._unique_type_triples(return_counts=True)
+        metric_names = ["count", "min", "max", "mean", "std"]
+        et_metrics = self._edge_metrics_by_type_triple(metrics=metric_names)
+
         edge_types = sorted(
             (
-                (count, EdgeType(src_ty, rel_ty, tgt_ty))
-                for src_ty, rel_ty, tgt_ty, count in unique_ets
+                (metrics.count, EdgeType(*metrics.Index), metrics,)
+                for metrics in et_metrics.itertuples()
             ),
             reverse=True,
         )
 
         edges = separated(
-            [f"{str_edge_type(et)}: [{count}]" for count, et in edge_types],
+            [edge_type_info(et, metrics) for _, et, metrics in edge_types],
             limit=truncate,
             stringify=str,
             sep="\n    ",
@@ -960,13 +1145,12 @@ class StellarGraph:
         if nodes is None:
             selector = slice(None)
         else:
-            selector = np.isin(self._edges.sources, nodes) & np.isin(
-                self._edges.targets, nodes
+            node_ilocs = self._nodes.ids.to_iloc(nodes)
+            selector = np.isin(self._edges.sources, node_ilocs) & np.isin(
+                self._edges.targets, node_ilocs
             )
 
-        for n1, rel, n2 in self._unique_type_triples(
-            selector=selector, return_counts=False
-        ):
+        for n1, rel, n2 in self._unique_type_triples(selector=selector):
             edge_type_tri = EdgeType(n1, rel, n2)
             edge_types.add(edge_type_tri)
             graph_schema[n1].add(edge_type_tri)
@@ -989,16 +1173,27 @@ class StellarGraph:
             self.is_directed(), sorted(self.node_types), edge_types, schema
         )
 
-    def node_degrees(self) -> Mapping[Any, int]:
+    def node_degrees(self, use_ilocs=False) -> Mapping[Any, int]:
         """
         Obtains a map from node to node degree.
 
+        use_ilocs (bool): if True return :ref:`node ilocs <iloc-explanation>`
+
         Returns:
             The degree of each node.
         """
-        return self._edges.degrees()
+        degrees = self._edges.degrees()
+        if use_ilocs:
+            return degrees
+        node_ids = self.node_ilocs_to_ids(list(degrees.keys()))
+        return defaultdict(
+            int,
+            ((node_id, degree) for node_id, degree in zip(node_ids, degrees.values())),
+        )
 
-    def to_adjacency_matrix(self, nodes: Optional[Iterable] = None, weighted=False):
+    def to_adjacency_matrix(
+        self, nodes: Optional[Iterable] = None, weighted=False, edge_type=None
+    ):
         """
         Obtains a SciPy sparse adjacency matrix of edge weights.
 
@@ -1012,36 +1207,48 @@ class StellarGraph:
                 otherwise, it is computed for the full graph.
             weighted (bool): If true, use the edge weight column from the graph instead
                 of edge counts (weights from multi-edges are summed).
+            edge_type (hashable, optional): If set (to an edge type), only includes edges of that
+                type, otherwise uses all edges.
 
         Returns:
              The weighted adjacency matrix.
         """
-        if nodes is None:
-            index = self._nodes._id_index
-            selector = slice(None)
+        if edge_type is None:
+            type_selector = slice(None)
         else:
-            nodes = list(nodes)
-            index = ExternalIdIndex(nodes)
-            selector = np.isin(self._edges.sources, nodes) & np.isin(
-                self._edges.targets, nodes
-            )
+            type_selector = self._edges.type_range(edge_type)
+
+        sources = self._edges.sources[type_selector]
+        targets = self._edges.targets[type_selector]
+
+        if nodes is None:
+            # if `nodes` is None use overall ilocs (for the original graph)
+            src_idx = sources
+            tgt_idx = targets
+            selector = slice(None)
+            n = self.number_of_nodes()
+        else:
+            node_ilocs = self._nodes.ids.to_iloc(nodes)
+            index = ExternalIdIndex(node_ilocs)
+            n = len(index)
+            selector = np.isin(sources, node_ilocs) & np.isin(targets, node_ilocs)
+
+            # these indices are computed relative to the index above
+            src_idx = index.to_iloc(sources[selector])
+            tgt_idx = index.to_iloc(targets[selector])
 
-        # these indices are computed relative to the index above. If `nodes` is None, they'll be the
-        # overall ilocs (for the original graph), otherwise they'll be the indices of the `nodes`
-        # list.
-        src_idx = index.to_iloc(self._edges.sources[selector])
-        tgt_idx = index.to_iloc(self._edges.targets[selector])
         if weighted:
-            weights = self._edges.weights[selector]
+            weights = self._edges.weights[type_selector][selector]
         else:
             weights = np.ones(src_idx.shape, dtype=self._edges.weights.dtype)
 
-        n = len(index)
-
         adj = sps.csr_matrix((weights, (src_idx, tgt_idx)), shape=(n, n))
-        if not self.is_directed():
+        if not self.is_directed() and n > 0:
             # in an undirected graph, the adjacency matrix should be symmetric: which means counting
             # weights from either "incoming" or "outgoing" edges, but not double-counting self loops
+
+            # FIXME https://github.com/scipy/scipy/issues/11949: these operations, particularly the
+            # diagonal, don't work for an empty matrix (n == 0)
             backward = adj.transpose(copy=True)
             # this is setdiag(0), but faster, since it doesn't change the sparsity structure of the
             # matrix (https://github.com/scipy/scipy/issues/11600)
@@ -1082,13 +1289,18 @@ class StellarGraph:
         # FIXME(#985): this is O(edges in graph) but could potentially be optimised to O(edges in
         # graph incident to `nodes`), which could be much fewer if `nodes` is small
         edge_ilocs = np.where(
-            np.isin(self._edges.sources, nodes) & np.isin(self._edges.targets, nodes)
+            np.isin(self._edges.sources, node_ilocs)
+            & np.isin(self._edges.targets, node_ilocs)
         )
         edge_frame = pd.DataFrame(
             {
                 "id": self._edges.ids.from_iloc(edge_ilocs),
-                globalvar.SOURCE: self._edges.sources[edge_ilocs],
-                globalvar.TARGET: self._edges.targets[edge_ilocs],
+                globalvar.SOURCE: self._nodes.ids.from_iloc(
+                    self._edges.sources[edge_ilocs]
+                ),
+                globalvar.TARGET: self._nodes.ids.from_iloc(
+                    self._edges.targets[edge_ilocs]
+                ),
                 globalvar.WEIGHT: self._edges.weights[edge_ilocs],
             },
             index=self._edges.type_of_iloc(edge_ilocs),
@@ -1213,8 +1425,8 @@ class StellarGraph:
                 graph.add_nodes_from(node_ids, **ty_dict)
 
         iterator = zip(
-            self._edges.sources,
-            self._edges.targets,
+            self._nodes.ids.from_iloc(self._edges.sources),
+            self._nodes.ids.from_iloc(self._edges.targets),
             self._edges.type_of_iloc(slice(None)),
             self._edges.weights,
         )
@@ -1225,25 +1437,7 @@ class StellarGraph:
 
         return graph
 
-    # FIXME: Experimental/special-case methods that need to be considered more; the underscores
-    # denote "package private", not fully private, and so are ok to use in the rest of stellargraph
-    def _get_index_for_nodes(self, nodes, node_type=None):
-        """
-        Get the indices for the specified node or nodes.
-        If the node type is not specified the node types will be found
-        for all nodes. It is therefore important to supply the ``node_type``
-        for this method to be fast.
-
-        Args:
-            n: (list or hashable) Node ID or list of node IDs
-            node_type: (hashable) the type of the nodes.
-
-        Returns:
-            Numpy array containing the indices for the requested nodes.
-        """
-        return self._nodes._id_index.to_iloc(nodes, strict=True)
-
-    def _adjacency_types(self, graph_schema: GraphSchema):
+    def _adjacency_types(self, graph_schema: GraphSchema, use_ilocs=False):
         """
         Obtains the edges in the form of the typed mapping:
 
@@ -1251,6 +1445,7 @@ class StellarGraph:
 
         Args:
             graph_schema: The graph schema.
+            use_ilocs (bool): if True return :ref:`node ilocs <iloc-explanation>`
         Returns:
              The edge types mapping.
         """
@@ -1258,13 +1453,13 @@ class StellarGraph:
 
         triples = defaultdict(lambda: defaultdict(lambda: []))
 
-        iterator = zip(
-            source_types,
-            rel_types,
-            target_types,
-            self._edges.sources,
-            self._edges.targets,
-        )
+        sources = self._edges.sources
+        targets = self._edges.targets
+        if not use_ilocs:
+            sources = self._nodes.ids.from_iloc(sources)
+            targets = self._nodes.ids.from_iloc(targets)
+
+        iterator = zip(source_types, rel_types, target_types, sources, targets,)
         for src_type, rel_type, tgt_type, src, tgt in iterator:
             triple = EdgeType(src_type, rel_type, tgt_type)
             triples[triple][src].append(tgt)
@@ -1280,13 +1475,16 @@ class StellarGraph:
 
         return triples
 
-    def _edge_weights(self, source_node: Any, target_node: Any) -> List[Any]:
+    def _edge_weights(
+        self, source_node: Any, target_node: Any, use_ilocs=False
+    ) -> List[Any]:
         """
         Obtains the weights of edges between the given pair of nodes.
 
         Args:
-            source_node (any): The source node.
-            target_node (any): The target node.
+            source_node (int): The source node.
+            target_node (int): The target node.
+            use_ilocs (bool): if True source_node and target_node are treated as :ref:`node ilocs <iloc-explanation>`.
 
         Returns:
             list: The edge weights.
@@ -1299,10 +1497,18 @@ class StellarGraph:
         effectively_directed = self.is_directed() or source_node == target_node
         both_dirs = not effectively_directed
 
-        source_ilocs = self._edges.edge_ilocs(source_node, ins=both_dirs, outs=True)
-        target_ilocs = self._edges.edge_ilocs(target_node, ins=True, outs=both_dirs)
+        if not use_ilocs:
+            source_node = self._nodes.ids.to_iloc([source_node])[0]
+            target_node = self._nodes.ids.to_iloc([target_node])[0]
 
-        ilocs = np.intersect1d(source_ilocs, target_ilocs, assume_unique=True)
+        source_edge_ilocs = self._edges.edge_ilocs(
+            source_node, ins=both_dirs, outs=True
+        )
+        target_edge_ilocs = self._edges.edge_ilocs(
+            target_node, ins=True, outs=both_dirs
+        )
+
+        ilocs = np.intersect1d(source_edge_ilocs, target_edge_ilocs, assume_unique=True)
 
         return [float(x) for x in self._edges.weights[ilocs]]
 
@@ -1317,6 +1523,7 @@ class StellarDiGraph(StellarGraph):
         source_column=globalvar.SOURCE,
         target_column=globalvar.TARGET,
         edge_weight_column=globalvar.WEIGHT,
+        edge_type_column=None,
         node_type_default=globalvar.NODE_TYPE_DEFAULT,
         edge_type_default=globalvar.EDGE_TYPE_DEFAULT,
         dtype="float32",
@@ -1333,6 +1540,7 @@ class StellarDiGraph(StellarGraph):
             source_column=source_column,
             target_column=target_column,
             edge_weight_column=edge_weight_column,
+            edge_type_column=edge_type_column,
             node_type_default=node_type_default,
             edge_type_default=edge_type_default,
             dtype=dtype,
