@@ -34,8 +34,67 @@ from scipy.special import softmax
 from ..core.schema import GraphSchema
 from ..core.graph import StellarGraph
 from ..core.utils import is_real_iterable
-from ..core.experimental import experimental
+from ..core.validation import require_integer_in_range, comma_sep
 from ..random import random_state
+from abc import ABC, abstractmethod
+
+
+def _default_if_none(value, default, name, ensure_not_none=True):
+    value = value if value is not None else default
+    if ensure_not_none and value is None:
+        raise ValueError(
+            f"{name}: expected a value to be specified in either `__init__` or `run`, found None in both"
+        )
+    return value
+
+
+class RandomWalk(ABC):
+    """
+    Abstract base class for Random Walk classes. A Random Walk class must implement a ``run`` method
+    which takes an iterable of node IDs and returns a list of walks. Each walk is a list of node IDs
+    that contains the starting node as its first element.
+    """
+
+    def __init__(self, graph, seed=None):
+        if not isinstance(graph, StellarGraph):
+            raise TypeError("Graph must be a StellarGraph or StellarDiGraph.")
+
+        self.graph = graph
+        self._random_state, self._np_random_state = random_state(seed)
+
+    def _get_random_state(self, seed):
+        """
+        Args:
+            seed: The optional seed value for a given run.
+
+        Returns:
+            The random state as determined by the seed.
+        """
+        if seed is None:
+            # Restore the random state
+            return self._random_state
+        # seed the random number generator
+        require_integer_in_range(seed, "seed", min_val=0)
+        rs, _ = random_state(seed)
+        return rs
+
+    @staticmethod
+    def _validate_walk_params(nodes, n, length):
+        if not is_real_iterable(nodes):
+            raise ValueError(f"nodes: expected an iterable, found: {nodes}")
+        if len(nodes) == 0:
+            warnings.warn(
+                "No root node IDs given. An empty list will be returned as a result.",
+                RuntimeWarning,
+                stacklevel=3,
+            )
+
+        require_integer_in_range(n, "n", min_val=1)
+        require_integer_in_range(length, "length", min_val=1)
+
+    @abstractmethod
+    def run(self, nodes, **kwargs):
+        pass
 
 
 class GraphWalk(object):
@@ -48,7 +107,6 @@ class GraphWalk(object):
 
         # Initialize the random state
         self._check_seed(seed)
-
         self._random_state, self._np_random_state = random_state(seed)
 
         # We require a StellarGraph for this
@@ -70,7 +128,9 @@ class GraphWalk(object):
         adj = getattr(self, "adj_types", None)
         if not adj:
             # Create a dict of adjacency lists per edge type, for faster neighbour sampling from graph in SampledHeteroBFS:
-            self.adj_types = adj = self.graph._adjacency_types(self.graph_schema)
+            self.adj_types = adj = self.graph._adjacency_types(
+                self.graph_schema, use_ilocs=True
+            )
         return adj
 
     def _check_seed(self, seed):
@@ -93,16 +153,14 @@ class GraphWalk(object):
             The random state as determined by the seed.
         """
         if seed is None:
-            # Restore the random state
+            # Use the class's random state
             return self._random_state
         # seed the random number generator
         rs, _ = random_state(seed)
         return rs
 
     def neighbors(self, node):
-        if not self.graph.has_node(node):
-            self._raise_error("node {} not in graph".format(node))
-        return self.graph.neighbors(node)
+        return self.graph.neighbors(node, use_ilocs=True)
 
     def run(self, *args, **kwargs):
         """
@@ -176,28 +234,45 @@ class GraphWalk(object):
                 self._raise_error(err_msg)
 
 
-class UniformRandomWalk(GraphWalk):
+class UniformRandomWalk(RandomWalk):
     """
     Performs uniform random walks on the given graph
+
+    Args:
+        graph (StellarGraph): Graph to traverse
+        n (int, optional): Total number of random walks per root node
+        length (int, optional): Maximum length of each random walk
+        seed (int, optional): Random number generator seed
+
     """
 
-    def run(self, nodes, n, length, seed=None):
+    def __init__(self, graph, n=None, length=None, seed=None):
+        super().__init__(graph, seed=seed)
+        self.n = n
+        self.length = length
+
+    def run(self, nodes, *, n=None, length=None, seed=None):
         """
-        Perform a random walk starting from the root nodes.
+        Perform a random walk starting from the root nodes. Optional parameters default to using the
+        values passed in during construction.
 
         Args:
             nodes (list): The root nodes as a list of node IDs
-            n (int): Total number of random walks per root node
-            length (int): Maximum length of each random walk
-            seed (int, optional): Random number generator seed; default is None
+            n (int, optional): Total number of random walks per root node
+            length (int, optional): Maximum length of each random walk
+            seed (int, optional): Random number generator seed
 
         Returns:
             List of lists of nodes ids for each of the random walks
 
         """
-        self._check_common_parameters(nodes, n, length, seed)
+        n = _default_if_none(n, self.n, "n")
+        length = _default_if_none(length, self.length, "length")
+        self._validate_walk_params(nodes, n, length)
         rs = self._get_random_state(seed)
 
+        nodes = self.graph.node_ids_to_ilocs(nodes)
+
         # for each root node, do n walks
         return [self._walk(rs, node, length) for node in nodes for _ in range(n)]
 
@@ -205,7 +280,7 @@ class UniformRandomWalk(GraphWalk):
         walk = [start_node]
         current_node = start_node
         for _ in range(length - 1):
-            neighbours = self.neighbors(current_node)
+            neighbours = self.graph.neighbors(current_node, use_ilocs=True)
             if not neighbours:
                 # dead end, so stop
                 break
@@ -214,7 +289,7 @@ class UniformRandomWalk(GraphWalk):
                 current_node = rs.choice(neighbours)
             walk.append(current_node)
 
-        return walk
+        return list(self.graph.node_ilocs_to_ids(walk))
 
 
 def naive_weighted_choices(rs, weights):
@@ -252,71 +327,114 @@ def naive_weighted_choices(rs, weights):
     return idx
 
 
-class BiasedRandomWalk(GraphWalk):
+class BiasedRandomWalk(RandomWalk):
     """
     Performs biased second order random walks (like those used in Node2Vec algorithm
     https://snap.stanford.edu/node2vec/) controlled by the values of two parameters p and q.
+
+    Args:
+        graph (StellarGraph): Graph to traverse
+        n (int, optional): Total number of random walks per root node
+        length (int, optional): Maximum length of each random walk
+        p (float, optional): Defines probability, 1/p, of returning to source node
+        q (float, optional): Defines probability, 1/q, for moving to a node away from the source node
+        weighted (bool, optional): Indicates whether the walk is unweighted or weighted
+        seed (int, optional): Random number generator seed
+
     """
 
-    def run(self, nodes, n, length, p=1.0, q=1.0, seed=None, weighted=False):
+    def __init__(
+        self, graph, n=None, length=None, p=1.0, q=1.0, weighted=False, seed=None,
+    ):
+        super().__init__(graph, seed=seed)
+        self.n = n
+        self.length = length
+        self.p = p
+        self.q = q
+        self.weighted = weighted
+        self._checked_weights = False
+
+        if weighted:
+            self._check_weights_valid()
+
+    def _check_weights_valid(self):
+        if self._checked_weights:
+            # we only need to check the weights once, either in the constructor or in run, whichever
+            # sets `weighted=True` first
+            return
+
+        # Check that all edge weights are greater than or equal to 0.
+        edges, weights = self.graph.edges(include_edge_weight=True, use_ilocs=True)
+        (invalid,) = np.where((weights < 0) | ~np.isfinite(weights))
+        if len(invalid) > 0:
+
+            def format(idx):
+                s, t = edges[idx]
+                w = weights[idx]
+                return f"{s!r} to {t!r} (weight = {w})"
+
+            raise ValueError(
+                f"graph: expected all edge weights to be non-negative and finite, found some negative or infinite: {comma_sep(invalid, stringify=format)}"
+            )
+
+        self._checked_weights = True
+
+    def run(
+        self, nodes, *, n=None, length=None, p=None, q=None, seed=None, weighted=None
+    ):
 
         """
-        Perform a random walk starting from the root nodes.
+        Perform a random walk starting from the root nodes. Optional parameters default to using the
+        values passed in during construction.
 
         Args:
             nodes (list): The root nodes as a list of node IDs
-            n (int): Total number of random walks per root node
-            length (int): Maximum length of each random walk
-            p (float, default 1.0): Defines probability, 1/p, of returning to source node
-            q (float, default 1.0): Defines probability, 1/q, for moving to a node away from the source node
+            n (int, optional): Total number of random walks per root node
+            length (int, optional): Maximum length of each random walk
+            p (float, optional): Defines probability, 1/p, of returning to source node
+            q (float, optional): Defines probability, 1/q, for moving to a node away from the source node
             seed (int, optional): Random number generator seed; default is None
-            weighted (bool, default False): Indicates whether the walk is unweighted or weighted
+            weighted (bool, optional): Indicates whether the walk is unweighted or weighted
 
         Returns:
             List of lists of nodes ids for each of the random walks
 
         """
-        self._check_common_parameters(nodes, n, length, seed)
+        n = _default_if_none(n, self.n, "n")
+        length = _default_if_none(length, self.length, "length")
+        p = _default_if_none(p, self.p, "p")
+        q = _default_if_none(q, self.q, "q")
+        weighted = _default_if_none(weighted, self.weighted, "weighted")
+        self._validate_walk_params(nodes, n, length)
         self._check_weights(p, q, weighted)
         rs = self._get_random_state(seed)
 
+        nodes = self.graph.node_ids_to_ilocs(nodes)
+
         if weighted:
-            # Check that all edge weights are greater than or equal to 0.
-            # Also, if the given graph is a MultiGraph, then check that there are no two edges between
-            # the same two nodes with different weights.
-            for node in self.graph.nodes():
-                # TODO Encapsulate edge weights
-                for neighbor in self.graph.neighbors(node):
+            self._check_weights_valid()
 
-                    wts = set()
-                    for weight in self.graph._edge_weights(node, neighbor):
-                        if weight is None or np.isnan(weight) or weight == np.inf:
-                            self._raise_error(
-                                "Missing or invalid edge weight ({}) between ({}) and ({}).".format(
-                                    weight, node, neighbor
-                                )
-                            )
-                        if not isinstance(weight, (int, float)):
-                            self._raise_error(
-                                "Edge weight between nodes ({}) and ({}) is not numeric ({}).".format(
-                                    node, neighbor, weight
-                                )
-                            )
-                        if weight < 0:  # check if edge has a negative weight
-                            self._raise_error(
-                                "An edge weight between nodes ({}) and ({}) is negative ({}).".format(
-                                    node, neighbor, weight
-                                )
-                            )
+            # calculate the appropriate unnormalised transition probability, given the history of
+            # the walk
+            def transition_probability(nn):
+                nn, weight = nn
 
-                        wts.add(weight)
-                    if len(wts) > 1:
-                        # multigraph with different weights on edges between same pair of nodes
-                        self._raise_error(
-                            "({}) and ({}) have multiple edges with weights ({}). Ambiguous to choose an edge for the random walk.".format(
-                                node, neighbor, list(wts)
-                            )
-                        )
+                if nn == previous_node:  # d_tx = 0
+                    return ip * weight
+                elif any(nn == pn for pn, _ in previous_node_neighbours):  # d_tx = 1
+                    return weight
+                else:  # d_tx = 2
+                    return iq * weight
+
+        else:
+            # without weights
+            def transition_probability(nn):
+                if nn == previous_node:  ## d_tx = 0
+                    return ip
+                elif nn in previous_node_neighbours:  # d_tx = 1
+                    return 1.0
+                else:  # d_tx = 2
+                    return iq
 
         ip = 1.0 / p
         iq = 1.0 / q
@@ -327,52 +445,34 @@ class BiasedRandomWalk(GraphWalk):
                 # the walk starts at the root
                 walk = [node]
 
-                neighbours = self.neighbors(node)
+                previous_node = None
+                previous_node_neighbours = []
 
-                previous_node = node
-                previous_node_neighbours = neighbours
+                current_node = node
 
-                # calculate the appropriate unnormalised transition
-                # probability, given the history of the walk
-                def transition_probability(nn, current_node, weighted):
+                for _ in range(length - 1):
+                    neighbours = self.graph.neighbors(
+                        current_node, use_ilocs=True, include_edge_weight=weighted
+                    )
 
+                    if not neighbours:
+                        break
+
+                    # select one of the neighbours using the
+                    # appropriate transition probabilities
+                    choice = naive_weighted_choices(
+                        rs, (transition_probability(nn) for nn in neighbours),
+                    )
+
+                    previous_node = current_node
+                    previous_node_neighbours = neighbours
+                    current_node = neighbours[choice]
                     if weighted:
-                        # TODO Encapsulate edge weights
-                        weight_cn = self.graph._edge_weights(current_node, nn)[0]
-                    else:
-                        weight_cn = 1.0
+                        current_node = current_node.node
 
-                    if nn == previous_node:  # d_tx = 0
-                        return ip * weight_cn
-                    elif nn in previous_node_neighbours:  # d_tx = 1
-                        return 1.0 * weight_cn
-                    else:  # d_tx = 2
-                        return iq * weight_cn
+                    walk.append(current_node)
 
-                if neighbours:
-                    current_node = rs.choice(neighbours)
-                    for _ in range(length - 1):
-                        walk.append(current_node)
-                        neighbours = self.neighbors(current_node)
-
-                        if not neighbours:
-                            break
-
-                        # select one of the neighbours using the
-                        # appropriate transition probabilities
-                        choice = naive_weighted_choices(
-                            rs,
-                            (
-                                transition_probability(nn, current_node, weighted)
-                                for nn in neighbours
-                            ),
-                        )
-
-                        previous_node = current_node
-                        previous_node_neighbours = neighbours
-                        current_node = neighbours[choice]
-
-                walks.append(walk)
+                walks.append(list(self.graph.node_ilocs_to_ids(walk)))
 
         return walks
 
@@ -387,31 +487,48 @@ class BiasedRandomWalk(GraphWalk):
             weighted: <False or True> Indicates whether the walk is unweighted or weighted.
        """
         if p <= 0.0:
-            self._raise_error("Parameter p should be greater than 0.")
+            raise ValueError(f"p: expected positive numeric value, found {p}")
 
         if q <= 0.0:
-            self._raise_error("Parameter q should be greater than 0.")
+            raise ValueError(f"q: expected positive numeric value, found {q}")
 
         if type(weighted) != bool:
-            self._raise_error(
-                "Parameter weighted has to be either False (unweighted random walks) or True (weighted random walks)."
-            )
+            raise ValueError(f"weighted: expected boolean value, found {weighted}")
 
 
-class UniformRandomMetaPathWalk(GraphWalk):
+class UniformRandomMetaPathWalk(RandomWalk):
     """
-    For heterogeneous graphs, it performs uniform random walks based on given metapaths.
+    For heterogeneous graphs, it performs uniform random walks based on given metapaths. Optional
+    parameters default to using the values passed in during construction.
+
+    Args:
+        graph (StellarGraph): Graph to traverse
+        n (int, optional): Total number of random walks per root node
+        length (int, optional): Maximum length of each random walk
+        metapaths (list of list, optional): List of lists of node labels that specify a metapath schema, e.g.,
+            [['Author', 'Paper', 'Author'], ['Author, 'Paper', 'Venue', 'Paper', 'Author']] specifies two metapath
+            schemas of length 3 and 5 respectively.
+        seed (int, optional): Random number generator seed
+
     """
 
-    def run(self, nodes, n, length, metapaths, seed=None):
+    def __init__(
+        self, graph, n=None, length=None, metapaths=None, seed=None,
+    ):
+        super().__init__(graph, seed=seed)
+        self.n = n
+        self.length = length
+        self.metapaths = metapaths
+
+    def run(self, nodes, *, n=None, length=None, metapaths=None, seed=None):
         """
         Performs metapath-driven uniform random walks on heterogeneous graphs.
 
         Args:
             nodes (list): The root nodes as a list of node IDs
-            n (int): Total number of random walks per root node
-            length (int): Maximum length of each random walk
-            metapaths (list of list): List of lists of node labels that specify a metapath schema, e.g.,
+            n (int, optional): Total number of random walks per root node
+            length (int, optional): Maximum length of each random walk
+            metapaths (list of list, optional): List of lists of node labels that specify a metapath schema, e.g.,
                 [['Author', 'Paper', 'Author'], ['Author, 'Paper', 'Venue', 'Paper', 'Author']] specifies two metapath
                 schemas of length 3 and 5 respectively.
             seed (int, optional): Random number generator seed; default is None
@@ -419,15 +536,20 @@ class UniformRandomMetaPathWalk(GraphWalk):
         Returns:
             List of lists of nodes ids for each of the random walks generated
         """
-        self._check_common_parameters(nodes, n, length, seed)
+        n = _default_if_none(n, self.n, "n")
+        length = _default_if_none(length, self.length, "length")
+        metapaths = _default_if_none(metapaths, self.metapaths, "metapaths")
+        self._validate_walk_params(nodes, n, length)
         self._check_metapath_values(metapaths)
         rs = self._get_random_state(seed)
 
+        nodes = self.graph.node_ids_to_ilocs(nodes)
+
         walks = []
 
         for node in nodes:
             # retrieve node type
-            label = self.graph.node_type(node)
+            label = self.graph.node_type(node, use_ilocs=True)
             filtered_metapaths = [
                 metapath
                 for metapath in metapaths
@@ -450,12 +572,13 @@ class UniformRandomMetaPathWalk(GraphWalk):
                     for d in range(length):
                         walk.append(current_node)
                         # d+1 can also be used to index metapath to retrieve the node type for the next step in the walk
-                        neighbours = self.neighbors(current_node)
+                        neighbours = self.graph.neighbors(current_node, use_ilocs=True)
                         # filter these by node type
                         neighbours = [
                             n_node
                             for n_node in neighbours
-                            if self.graph.node_type(n_node) == metapath[d]
+                            if self.graph.node_type(n_node, use_ilocs=True)
+                            == metapath[d]
                         ]
                         if len(neighbours) == 0:
                             # if no neighbours of the required type as dictated by the metapath exist, then stop.
@@ -465,7 +588,9 @@ class UniformRandomMetaPathWalk(GraphWalk):
                             neighbours
                         )  # the next node in the walk
 
-                    walks.append(walk)  # store the walk
+                    walks.append(
+                        list(self.graph.node_ilocs_to_ids(walk))
+                    )  # store the walk
 
         return walks
 
@@ -479,20 +604,24 @@ class UniformRandomMetaPathWalk(GraphWalk):
                 [['Author', 'Paper', 'Author'], ['Author, 'Paper', 'Venue', 'Paper', 'Author']] specifies two metapath
                 schemas of length 3 and 5 respectively.
         """
+
+        def raise_error(msg):
+            raise ValueError(f"metapaths: {msg}, found {metapaths}")
+
         if type(metapaths) != list:
-            self._raise_error("The metapaths parameter must be a list of lists.")
+            raise_error("expected list of lists.")
         for metapath in metapaths:
             if type(metapath) != list:
-                self._raise_error("Each metapath must be list type of node labels")
+                raise_error("expected each metapath to be a list of node labels")
             if len(metapath) < 2:
-                self._raise_error("Each metapath must specify at least two node types")
+                raise_error("expected each metapath to specify at least two node types")
 
             for node_label in metapath:
                 if type(node_label) != str:
-                    self._raise_error("Node labels in metapaths must be string type.")
+                    raise_error("expected each node type in metapaths to be a string")
             if metapath[0] != metapath[-1]:
-                self._raise_error(
-                    "The first and last node type in a metapath should be the same."
+                raise_error(
+                    "expected the first and last node type in a metapath to be the same"
                 )
 
 
@@ -543,12 +672,14 @@ class SampledBreadthFirstWalk(GraphWalk):
                     # consider the subgraph up to and including max_hops from root node
                     if depth > max_hops:
                         continue
+
                     neighbours = (
                         self.neighbors(cur_node) if cur_node is not None else []
                     )
                     if len(neighbours) == 0:
                         # Either node is unconnected or is in directed graph with no out-nodes.
-                        neighbours = [None] * n_size[cur_depth]
+                        _size = n_size[cur_depth]
+                        neighbours = [-1] * _size
                     else:
                         # sample with replacement
                         neighbours = rs.choices(neighbours, k=n_size[cur_depth])
@@ -599,7 +730,7 @@ class SampledHeterogeneousBreadthFirstWalk(GraphWalk):
                 walk = list()  # the list of nodes in the subgraph of node
 
                 # Start the walk by adding the head node, and node type to the frontier list q
-                node_type = self.graph.node_type(node)
+                node_type = self.graph.node_type(node, use_ilocs=True)
                 q.extend([(node, node_type, 0)])
 
                 # add the root node to the walks
@@ -627,7 +758,8 @@ class SampledHeterogeneousBreadthFirstWalk(GraphWalk):
                             if len(neigh_et) > 0:
                                 samples = rs.choices(neigh_et, k=n_size[depth - 1])
                             else:  # this doesn't happen anymore, see the comment above
-                                samples = [None] * n_size[depth - 1]
+                                _size = n_size[depth - 1]
+                                samples = [-1] * _size
 
                             walk.append(samples)
                             q.extend(
@@ -757,13 +889,15 @@ class DirectedBreadthFirstNeighbours(GraphWalk):
         """
         if node is None:
             # Non-node, e.g. previously sampled from empty neighbourhood
-            return [None] * size
+            return [-1] * size
         neighbours = list(
-            self.graph.in_nodes(node) if idx == 0 else self.graph.out_nodes(node)
+            self.graph.in_nodes(node, use_ilocs=True)
+            if idx == 0
+            else self.graph.out_nodes(node, use_ilocs=True)
         )
         if len(neighbours) == 0:
             # Sampling from empty neighbourhood
-            return [None] * size
+            return [-1] * size
         # Sample with replacement
         return rs.choices(neighbours, k=size)
 
@@ -791,28 +925,74 @@ class TemporalRandomWalk(GraphWalk):
     Performs temporal random walks on the given graph. The graph should contain numerical edge
     weights that correspond to the time at which the edge was created. Exact units are not relevant
     for the algorithm, only the relative differences (e.g. seconds, days, etc).
+
+    Args:
+        graph (StellarGraph): Graph to traverse
+        cw_size (int, optional): Size of context window. Also used as the minimum walk length,
+            since a walk must generate at least 1 context window for it to be useful.
+        max_walk_length (int, optional): Maximum length of each random walk. Should be greater
+            than or equal to the context window size.
+        initial_edge_bias (str, optional): Distribution to use when choosing a random
+            initial temporal edge to start from. Available options are:
+
+            * None (default) - The initial edge is picked from a uniform distribution.
+            * "exponential" - Heavily biased towards more recent edges.
+
+        walk_bias (str, optional): Distribution to use when choosing a random
+            neighbour to walk through. Available options are:
+
+            * None (default) - Neighbours are picked from a uniform distribution.
+            * "exponential" - Exponentially decaying probability, resulting in a bias towards shorter time gaps.
+
+        p_walk_success_threshold (float, optional): Lower bound for the proportion of successful
+            (i.e. longer than minimum length) walks. If the 95% percentile of the
+            estimated proportion is less than the provided threshold, a RuntimeError
+            will be raised. The default value of 0.01 means an error is raised if less than 1%
+            of the attempted random walks are successful. This parameter exists to catch any
+            potential situation where too many unsuccessful walks can cause an infinite or very
+            slow loop.
+        seed (int, optional): Random number generator seed.
+
     """
 
-    def run(
+    def __init__(
         self,
-        num_cw,
-        cw_size,
+        graph,
+        cw_size=None,
         max_walk_length=80,
         initial_edge_bias=None,
         walk_bias=None,
         p_walk_success_threshold=0.01,
         seed=None,
+    ):
+        super().__init__(graph, graph_schema=None, seed=seed)
+        self.cw_size = cw_size
+        self.max_walk_length = max_walk_length
+        self.initial_edge_bias = initial_edge_bias
+        self.walk_bias = walk_bias
+        self.p_walk_success_threshold = p_walk_success_threshold
+
+    def run(
+        self,
+        num_cw,
+        cw_size=None,
+        max_walk_length=None,
+        initial_edge_bias=None,
+        walk_bias=None,
+        p_walk_success_threshold=None,
+        seed=None,
     ):
         """
         Perform a time respecting random walk starting from randomly selected temporal edges.
+        Optional parameters default to using the values passed in during construction.
 
         Args:
             num_cw (int): Total number of context windows to generate. For comparable
                 results to most other random walks, this should be a multiple of the number
                 of nodes in the graph.
-            cw_size (int): Size of context window. Also used as the minimum walk length,
+            cw_size (int, optional): Size of context window. Also used as the minimum walk length,
                 since a walk must generate at least 1 context window for it to be useful.
-            max_walk_length (int): Maximum length of each random walk. Should be greater
+            max_walk_length (int, optional): Maximum length of each random walk. Should be greater
                 than or equal to the context window size.
             initial_edge_bias (str, optional): Distribution to use when choosing a random
                 initial temporal edge to start from. Available options are:
@@ -826,7 +1006,7 @@ class TemporalRandomWalk(GraphWalk):
                 * None (default) - Neighbours are picked from a uniform distribution.
                 * "exponential" - Exponentially decaying probability, resulting in a bias towards shorter time gaps.
 
-            p_walk_success_threshold (float): Lower bound for the proportion of successful
+            p_walk_success_threshold (float, optional): Lower bound for the proportion of successful
                 (i.e. longer than minimum length) walks. If the 95% percentile of the
                 estimated proportion is less than the provided threshold, a RuntimeError
                 will be raised. The default value of 0.01 means an error is raised if less than 1%
@@ -839,6 +1019,25 @@ class TemporalRandomWalk(GraphWalk):
             List of lists of node ids for each of the random walks.
 
         """
+        cw_size = _default_if_none(cw_size, self.cw_size, "cw_size")
+        max_walk_length = _default_if_none(
+            max_walk_length, self.max_walk_length, "max_walk_length"
+        )
+        initial_edge_bias = _default_if_none(
+            initial_edge_bias,
+            self.initial_edge_bias,
+            "initial_edge_bias",
+            ensure_not_none=False,
+        )
+        walk_bias = _default_if_none(
+            walk_bias, self.walk_bias, "walk_bias", ensure_not_none=False
+        )
+        p_walk_success_threshold = _default_if_none(
+            p_walk_success_threshold,
+            self.p_walk_success_threshold,
+            "p_walk_success_threshold",
+        )
+
         if cw_size < 2:
             raise ValueError(
                 f"cw_size: context window size should be greater than 1, found {cw_size}"
