@@ -26,6 +26,13 @@ from stellargraph.data.explorer import UniformRandomWalk
 from stellargraph.random import random_state
 
 
+def _warn_if_ignored(value, default, name):
+    if value != default:
+        raise ValueError(
+            f"walker, {name}: cannot specify both 'walker' and '{name}'. Please use one or the other."
+        )
+
+
 class UnsupervisedSampler:
     """
         The UnsupervisedSampler is responsible for sampling walks in the given graph
@@ -34,18 +41,25 @@ class UnsupervisedSampler:
         The positive samples are all the (target, context) pairs from the walks and the negative
         samples are contexts generated for each target based on a sampling distribtution.
 
-        Currently only uniform random walks are performed, other walk strategies (such as
-        second order walks) will be enabled in the future.
+        By default, a UniformRandomWalk is used, but a custom `walker` can be specified instead. An
+        error will be raised if other parameters are specified along with a custom `walker`.
 
         Args:
             G (StellarGraph): A stellargraph with features.
-            nodes (optional, iterable) The root nodes from which individual walks start.
+            nodes (iterable, optional) The root nodes from which individual walks start.
                 If not provided, all nodes in the graph are used.
-            length (int): An integer giving the length of the walks. Length must be at least 2.
-            number_of_walks (int): Number of walks from each root node.
+            length (int): Length of the walks for the default UniformRandomWalk walker. Length must
+                be at least 2.
+            number_of_walks (int): Number of walks from each root node for the default
+                UniformRandomWalk walker.
+            seed (int, optional): Random seed for the default UniformRandomWalk walker.
+            walker (RandomWalk, optional): A RandomWalk object to use instead of the default
+                UniformRandomWalk walker.
     """
 
-    def __init__(self, G, nodes=None, length=2, number_of_walks=1, seed=None):
+    def __init__(
+        self, G, nodes=None, length=2, number_of_walks=1, seed=None, walker=None,
+    ):
         if not isinstance(G, StellarGraph):
             raise ValueError(
                 "({}) Graph must be a StellarGraph or StellarDigraph object.".format(
@@ -56,23 +70,15 @@ class UnsupervisedSampler:
             self.graph = G
 
         # Instantiate the walker class used to generate random walks in the graph
-        self.walker = UniformRandomWalk(G, seed=seed)
-
-        # This code will enable alternative walker classes
-        # TODO: Enable this code, but figure out how to pass required options to run
-        # if walker is not None:
-        #     if not isinstance(
-        #         walker, UniformRandomWalk
-        #     ):  # only work with Uniform Random Walker at the moment
-        #         raise TypeError(
-        #             "({}) Only Uniform Random Walks are possible".format(
-        #                 type(self).__name__
-        #             )
-        #         )
-        #     else:
-        #         self.walker = walker(G, seed=seed)
-        # else:
-        #         self.walker = UniformRandomWalk(G, seed=seed)
+        if walker is not None:
+            _warn_if_ignored(length, 2, "length")
+            _warn_if_ignored(number_of_walks, 1, "number_of_walks")
+            _warn_if_ignored(seed, None, "seed")
+            self.walker = walker
+        else:
+            self.walker = UniformRandomWalk(
+                G, n=number_of_walks, length=length, seed=seed
+            )
 
         # Define the root nodes for the walks
         # if no root nodes are provided for sampling defaulting to using all nodes as root nodes.
@@ -126,17 +132,15 @@ class UnsupervisedSampler:
         """
         self._check_parameter_values(batch_size)
 
-        all_nodes = list(self.graph.nodes())
+        all_nodes = list(self.graph.nodes(use_ilocs=True))
         # Use the sampling distribution as per node2vec
-        degrees = self.graph.node_degrees()
+        degrees = self.graph.node_degrees(use_ilocs=True)
         sampling_distribution = np.array([degrees[n] ** 0.75 for n in all_nodes])
         sampling_distribution_norm = sampling_distribution / np.sum(
             sampling_distribution
         )
 
-        walks = self.walker.run(
-            nodes=self.nodes, length=self.length, n=self.number_of_walks
-        )
+        walks = self.walker.run(nodes=self.nodes)
 
         # first item in each walk is the target/head node
         targets = [walk[0] for walk in walks]
@@ -149,9 +153,14 @@ class UnsupervisedSampler:
             ]
         )
 
+        positive_pairs = self.graph.node_ids_to_ilocs(positive_pairs.flatten()).reshape(
+            positive_pairs.shape
+        )
+
         negative_samples = self.np_random.choice(
             all_nodes, size=len(positive_pairs), p=sampling_distribution_norm
         )
+
         negative_pairs = np.column_stack((positive_pairs[:, 0], negative_samples))
 
         pairs = np.concatenate((positive_pairs, negative_pairs), axis=0)
