@@ -15,12 +15,12 @@
 # limitations under the License.
 
 import warnings
-from tensorflow.keras.layers import Dense, Lambda, Dropout, Input, Layer
+from tensorflow.keras.layers import Dense, Lambda, Dropout, Input, Layer, InputLayer
 import tensorflow.keras.backend as K
 
 from ..mapper import FullBatchGenerator
 from .preprocessing_layer import GraphPreProcessingLayer
-from .misc import SqueezedSparseConversion, deprecated_model_function
+from .misc import SqueezedSparseConversion, deprecated_model_function, GatherIndices
 
 
 class APPNPPropagationLayer(Layer):
@@ -34,23 +34,15 @@ class APPNPPropagationLayer(Layer):
         Keras requires this batch dimension, and for full-batch methods
         we only have a single "batch".
 
-      - There are three inputs required, the node features, the output
-        indices (the nodes that are to be selected in the final layer)
+      - There are two inputs required, the node features,
         and the normalized graph Laplacian matrix
 
       - This class assumes that the normalized Laplacian matrix is passed as
         input to the Keras methods.
 
-      - The output indices are used when ``final_layer=True`` and the returned outputs
-        are the final-layer features for the nodes indexed by output indices.
-
-      - If ``final_layer=False`` all the node features are output in the same ordering as
-        given by the adjacency matrix.
-
     Args:
         units (int): dimensionality of output feature vectors
-        final_layer (bool): If False the layer returns output for all nodes,
-                            if True it returns the subset specified by the indices passed to it.
+        final_layer (bool): Deprecated, use ``tf.gather`` or :class:`GatherIndices`
         teleport_probability: "probability" of returning to the starting node in the propogation step as desribed  in
         the paper (alpha in the paper)
         input_dim (int, optional): the size of the input shape, if known.
@@ -61,7 +53,7 @@ class APPNPPropagationLayer(Layer):
         self,
         units,
         teleport_probability=0.1,
-        final_layer=False,
+        final_layer=None,
         input_dim=None,
         **kwargs
     ):
@@ -72,7 +64,10 @@ class APPNPPropagationLayer(Layer):
 
         self.units = units
         self.teleport_probability = teleport_probability
-        self.final_layer = final_layer
+        if final_layer is not None:
+            raise ValueError(
+                "'final_layer' is not longer supported, use 'tf.gather' or 'GatherIndices' separately"
+            )
 
     def get_config(self):
         """
@@ -104,13 +99,10 @@ class APPNPPropagationLayer(Layer):
         Returns:
             An input shape tuple.
         """
-        feature_shape, out_shape, *As_shapes = input_shapes
+        feature_shape, *As_shapes = input_shapes
 
         batch_dim = feature_shape[0]
-        if self.final_layer:
-            out_dim = out_shape[1]
-        else:
-            out_dim = feature_shape[1]
+        out_dim = feature_shape[1]
 
         return batch_dim, out_dim, self.units
 
@@ -131,7 +123,6 @@ class APPNPPropagationLayer(Layer):
             inputs (list): a list of 3 input tensors that includes
                 propagated node features (size 1 x N x F),
                 node features (size 1 x N x F),
-                output indices (size 1 x M)
                 graph adjacency matrix (size N x N),
                 where N is the number of nodes in the graph, and
                 F is the dimensionality of node features.
@@ -139,7 +130,7 @@ class APPNPPropagationLayer(Layer):
         Returns:
             Keras Tensor that represents the output of the layer.
         """
-        propagated_features, features, out_indices, *As = inputs
+        propagated_features, features, *As = inputs
         batch_dim, n_nodes, _ = K.int_shape(features)
         if batch_dim != 1:
             raise ValueError(
@@ -149,7 +140,6 @@ class APPNPPropagationLayer(Layer):
         # Remove singleton batch dimension
         features = K.squeeze(features, 0)
         propagated_features = K.squeeze(propagated_features, 0)
-        out_indices = K.squeeze(out_indices, 0)
 
         # Propagate the node features
         A = As[0]
@@ -157,10 +147,6 @@ class APPNPPropagationLayer(Layer):
             A, propagated_features
         ) + self.teleport_probability * features
 
-        # On the final layer we gather the nodes referenced by the indices
-        if self.final_layer:
-            output = K.gather(output, out_indices)
-
         # Add batch dimension back if we removed it
         if batch_dim == 1:
             output = K.expand_dims(output, 0)
@@ -273,15 +259,14 @@ class APPNP:
         if self.method == "none":
             self.graph_norm_layer = GraphPreProcessingLayer(num_of_nodes=self.n_nodes)
 
-        self._layers = []
-
+        self._feature_layers = []
         # Initialize a stack of fully connected layers
         n_layers = len(self.layer_sizes)
         for ii in range(n_layers):
             l = self.layer_sizes[ii]
             a = self.activations[ii]
-            self._layers.append(Dropout(self.dropout))
-            self._layers.append(
+            self._feature_layers.append(Dropout(self.dropout))
+            self._feature_layers.append(
                 Dense(
                     l,
                     activation=a,
@@ -290,34 +275,17 @@ class APPNP:
                 )
             )
 
+        self._propagate_layers = []
         feature_dim = self.layer_sizes[-1]
         for ii in range(approx_iter):
-            self._layers.append(Dropout(self.dropout))
-            self._layers.append(
+            self._propagate_layers.append(Dropout(self.dropout))
+            self._propagate_layers.append(
                 APPNPPropagationLayer(
-                    feature_dim,
-                    teleport_probability=self.teleport_probability,
-                    final_layer=(ii == (self.approx_iter - 1)),
+                    feature_dim, teleport_probability=self.teleport_probability,
                 )
             )
 
-    def __call__(self, x):
-        """
-        Apply APPNP to the inputs.
-        The input tensors are expected to be a list of the following:
-        [
-            Node features shape (1, N, F),
-            Adjacency indices (1, E, 2),
-            Adjacency values (1, E),
-            Output indices (1, O)
-        ]
-        where N is the number of nodes, F the number of input features,
-              E is the number of edges, O the number of output nodes.
-        Args:
-            x (Tensor): input tensors
-        Returns:
-            Output tensor
-        """
+    def _run(self, x, feature_layers):
         x_in, out_indices, *As = x
 
         # Currently we require the batch dimension to be one for full-batch methods
@@ -348,28 +316,43 @@ class APPNP:
 
         h_layer = x_in
 
-        for layer in self._layers[: (2 * len(self.layer_sizes))]:
+        for layer in feature_layers:
             h_layer = layer(h_layer)
 
         feature_layer = h_layer
 
-        for layer in self._layers[(2 * len(self.layer_sizes)) :]:
+        for layer in self._propagate_layers:
             if isinstance(layer, APPNPPropagationLayer):
-                h_layer = layer([h_layer, feature_layer, out_indices] + Ainput)
+                h_layer = layer([h_layer, feature_layer] + Ainput)
             else:
                 # For other (non-graph) layers only supply the input tensor
                 h_layer = layer(h_layer)
 
+        # only return data for the requested nodes
+        h_layer = GatherIndices(batch_dims=1)([h_layer, out_indices])
+
         return h_layer
 
-    def in_out_tensors(self, multiplicity=None):
+    def __call__(self, x):
         """
-        Builds an APPNP model for node or link prediction
-
+        Apply APPNP to the inputs.
+        The input tensors are expected to be a list of the following:
+        [
+            Node features shape (1, N, F),
+            Adjacency indices (1, E, 2),
+            Adjacency values (1, E),
+            Output indices (1, O)
+        ]
+        where N is the number of nodes, F the number of input features,
+              E is the number of edges, O the number of output nodes.
+        Args:
+            x (Tensor): input tensors
         Returns:
-            tuple: `(x_inp, x_out)`, where `x_inp` is a list of Keras/TensorFlow
-            input tensors for the model and `x_out` is a tensor of the model output.
+            Output tensor
         """
+        return self._run(x, feature_layers=self._feature_layers)
+
+    def _tensors(self, multiplicity, feature_layers):
         # Inputs for features
         x_t = Input(batch_shape=(1, self.n_nodes, self.n_features))
 
@@ -398,8 +381,21 @@ class APPNP:
         # TODO: Support multiple matrices
 
         x_inp = [x_t, out_indices_t] + A_placeholders
-        x_out = self(x_inp)
 
+        x_out = self._run(x_inp, feature_layers=feature_layers)
+        return x_inp, x_out
+
+    def in_out_tensors(self, multiplicity=None):
+        """
+        Builds an APPNP model for node or link prediction
+
+        Returns:
+            tuple: `(x_inp, x_out)`, where `x_inp` is a list of Keras/TensorFlow
+            input tensors for the model and `x_out` is a tensor of the model output.
+        """
+        x_inp, x_out = self._tensors(
+            multiplicity=multiplicity, feature_layers=self._feature_layers
+        )
         # Flatten output by removing singleton batch dimension
         if x_out.shape[0] == 1:
             self.x_out_flat = Lambda(lambda x: K.squeeze(x, 0))(x_out)
@@ -438,50 +434,10 @@ class APPNP:
                 "APPNP does not currently support propagating a link model"
             )
 
-        out_indices_t = Input(batch_shape=(1, None), dtype="int32")
-
-        if self.use_sparse:
-            # Placeholders for the sparse adjacency matrix
-            A_indices_t = Input(batch_shape=(1, None, 2), dtype="int64")
-            A_values_t = Input(batch_shape=(1, None))
-            A_placeholders = [A_indices_t, A_values_t]
-
-        else:
-            # Placeholders for the dense adjacency matrix
-            A_m = Input(batch_shape=(1, self.n_nodes, self.n_nodes))
-            A_placeholders = [A_m]
-
-        if self.use_sparse:
-            A_indices, A_values = A_placeholders
-            Ainput = [
-                SqueezedSparseConversion(
-                    shape=(self.n_nodes, self.n_nodes), dtype=A_values.dtype
-                )([A_indices, A_values])
-            ]
-
-        # Otherwise, create dense matrix from input tensor
-        else:
-            Ainput = [Lambda(lambda A: K.squeeze(A, 0))(A) for A in A_placeholders]
-
-        # Inputs for features & target indices
-        x_t = Input(batch_shape=(1, self.n_nodes, self.n_features))
-        x_inp = [x_t, out_indices_t] + A_placeholders
-
-        # pass the node features through the base model
-        feature_layer = x_t
-        for layer in base_model.layers[1:]:
-            feature_layer = layer(feature_layer)
-
-        h_layer = feature_layer
-        # iterate through APPNPPropagation layers
-        for layer in self._layers[(2 * len(self.layer_sizes)) :]:
-            if isinstance(layer, APPNPPropagationLayer):
-                h_layer = layer([h_layer, feature_layer, out_indices_t] + Ainput)
-            else:
-                h_layer = layer(h_layer)
-
-        x_out = h_layer
-        return x_inp, x_out
+        feature_layers = [
+            layer for layer in base_model.layers if not isinstance(layer, InputLayer)
+        ]
+        return self._tensors(multiplicity=1, feature_layers=feature_layers)
 
     node_model = deprecated_model_function(_node_model, "node_model")
     link_model = deprecated_model_function(_link_model, "link_model")
