@@ -26,7 +26,7 @@ from tensorflow.keras import activations, constraints, initializers, regularizer
 from tensorflow.keras.layers import Input, Layer, Dropout, LeakyReLU, Lambda, Reshape
 
 from ..mapper import FullBatchNodeGenerator, FullBatchGenerator
-from .misc import SqueezedSparseConversion, deprecated_model_function
+from .misc import SqueezedSparseConversion, deprecated_model_function, GatherIndices
 
 
 class GraphAttention(Layer):
@@ -42,19 +42,12 @@ class GraphAttention(Layer):
         Keras requires this batch dimension, and for full-batch methods
         we only have a single "batch".
 
-      - There are three inputs required, the node features, the output
-        indices (the nodes that are to be selected in the final layer)
+      - There are two inputs required, the node features,
         and the graph adjacency matrix
 
       - This does not add self loops to the adjacency matrix, you should preprocess
         the adjacency matrix to add self-loops
 
-      - The output indices are used when ``final_layer=True`` and the returned outputs
-        are the final-layer features for the nodes indexed by output indices.
-
-      - If ``final_layer=False`` all the node features are output in the same ordering as
-        given by the adjacency matrix.
-
     Args:
         F_out (int): dimensionality of output feature vectors
         attn_heads (int or list of int): number of attention heads
@@ -63,8 +56,7 @@ class GraphAttention(Layer):
         in_dropout_rate (float): dropout rate applied to features
         attn_dropout_rate (float): dropout rate applied to attention coefficients
         activation (str): nonlinear activation applied to layer's output to obtain output features (eq. 4 of the GAT paper)
-        final_layer (bool): If False the layer returns output for all nodes,
-                            if True it returns the subset specified by the indices passed to it.
+        final_layer (bool): Deprecated, use ``tf.gather`` or :class:`GatherIndices`
         use_bias (bool): toggles an optional bias
         saliency_map_support (bool): If calculating saliency maps using the tools in
             stellargraph.interpretability.saliency_maps this should be True. Otherwise this should be False (default).
@@ -88,7 +80,7 @@ class GraphAttention(Layer):
         attn_dropout_rate=0.0,
         activation="relu",
         use_bias=True,
-        final_layer=False,
+        final_layer=None,
         saliency_map_support=False,
         kernel_initializer="glorot_uniform",
         kernel_regularizer=None,
@@ -116,7 +108,10 @@ class GraphAttention(Layer):
         self.attn_dropout_rate = attn_dropout_rate  # dropout rate for attention coefs
         self.activation = activations.get(activation)  # Eq. 4 in the paper
         self.use_bias = use_bias
-        self.final_layer = final_layer
+        if final_layer is not None:
+            raise ValueError(
+                "'final_layer' is not longer supported, use 'tf.gather' or 'GatherIndices' separately"
+            )
 
         self.saliency_map_support = saliency_map_support
         # Populated by build()
@@ -156,7 +151,6 @@ class GraphAttention(Layer):
             "attn_dropout_rate": self.attn_dropout_rate,
             "activation": activations.serialize(self.activation),
             "use_bias": self.use_bias,
-            "final_layer": self.final_layer,
             "saliency_map_support": self.saliency_map_support,
             "kernel_initializer": initializers.serialize(self.kernel_initializer),
             "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
@@ -189,13 +183,10 @@ class GraphAttention(Layer):
         Returns:
             An input shape tuple.
         """
-        feature_shape, out_shape, *As_shapes = input_shapes
+        feature_shape, *As_shapes = input_shapes
 
         batch_dim = feature_shape[0]
-        if self.final_layer:
-            out_dim = out_shape[1]
-        else:
-            out_dim = feature_shape[1]
+        out_dim = feature_shape[1]
 
         return batch_dim, out_dim, self.output_dim
 
@@ -270,26 +261,22 @@ class GraphAttention(Layer):
         Keras requires this batch dimension, and for full-batch methods
         we only have a single "batch".
 
-        There are three inputs required, the node features, the output
-        indices (the nodes that are to be selected in the final layer)
+        There are two inputs required, the node features,
         and the graph adjacency matrix
 
         Notes:
             This does not add self loops to the adjacency matrix.
-            The output indices are only used when ``final_layer=True``
 
         Args:
             inputs (list): list of inputs with 3 items:
             node features (size 1 x N x F),
-            output indices (size 1 x M),
             graph adjacency matrix (size N x N),
             where N is the number of nodes in the graph,
                   F is the dimensionality of node features
                   M is the number of output nodes
         """
         X = inputs[0]  # Node features (1 x N x F)
-        out_indices = inputs[1]  # output indices (1 x K)
-        A = inputs[2]  # Adjacency matrix (N x N)
+        A = inputs[1]  # Adjacency matrix (N x N)
         N = K.int_shape(A)[-1]
 
         batch_dim, n_nodes, _ = K.int_shape(X)
@@ -301,7 +288,6 @@ class GraphAttention(Layer):
         else:
             # Remove singleton batch dimension
             X = K.squeeze(X, 0)
-            out_indices = K.squeeze(out_indices, 0)
 
         outputs = []
         for head in range(self.attn_heads):
@@ -373,10 +359,6 @@ class GraphAttention(Layer):
         # Nonlinear activation function
         output = self.activation(output)
 
-        # On the final layer we gather the nodes referenced by the indices
-        if self.final_layer:
-            output = K.gather(output, out_indices)
-
         # Add batch dimension back if we removed it
         if batch_dim == 1:
             output = K.expand_dims(output, 0)
@@ -403,13 +385,6 @@ class GraphAttentionSparse(GraphAttention):
       - This does not add self loops to the adjacency matrix, you should preprocess
         the adjacency matrix to add self-loops
 
-      - The output indices are used when `final_layer=True` and the returned outputs
-        are the final-layer features for the nodes indexed by output indices.
-
-      - If `final_layer=False` all the node features are output in the same ordering as
-        given by the adjacency matrix.
-
-
     Args:
         F_out (int): dimensionality of output feature vectors
         attn_heads (int or list of int): number of attention heads
@@ -418,8 +393,7 @@ class GraphAttentionSparse(GraphAttention):
         in_dropout_rate (float): dropout rate applied to features
         attn_dropout_rate (float): dropout rate applied to attention coefficients
         activation (str): nonlinear activation applied to layer's output to obtain output features (eq. 4 of the GAT paper)
-        final_layer (bool): If False the layer returns output for all nodes,
-                            if True it returns the subset specified by the indices passed to it.
+        final_layer (bool): Deprecated, use ``tf.gather`` or :class:`GatherIndices`
         use_bias (bool): toggles an optional bias
         saliency_map_support (bool): If calculating saliency maps using the tools in
             stellargraph.interpretability.saliency_maps this should be True. Otherwise this should be False (default).
@@ -440,20 +414,17 @@ class GraphAttentionSparse(GraphAttention):
 
         Notes:
             This does not add self loops to the adjacency matrix.
-            The output indices are only used when `final_layer=True`
 
         Args:
             inputs (list): list of inputs with 4 items:
             node features (size b x N x F),
-            output indices (size b x M),
             sparse graph adjacency matrix (size N x N),
             where N is the number of nodes in the graph,
                   F is the dimensionality of node features
                   M is the number of output nodes
         """
         X = inputs[0]  # Node features (1 x N x F)
-        out_indices = inputs[1]  # output indices (1 x K)
-        A_sparse = inputs[2]  # Adjacency matrix (1 x N x N)
+        A_sparse = inputs[1]  # Adjacency matrix (1 x N x N)
 
         if not isinstance(A_sparse, tf.SparseTensor):
             raise TypeError("A is not sparse")
@@ -468,7 +439,6 @@ class GraphAttentionSparse(GraphAttention):
             )
         else:
             # Remove singleton batch dimension
-            out_indices = K.squeeze(out_indices, 0)
             X = K.squeeze(X, 0)
 
         outputs = []
@@ -533,10 +503,6 @@ class GraphAttentionSparse(GraphAttention):
 
         output = self.activation(output)
 
-        # On the final layer we gather the nodes referenced by the indices
-        if self.final_layer:
-            output = K.gather(output, out_indices)
-
         # Add batch dimension back if we removed it
         if batch_dim == 1:
             output = K.expand_dims(output, 0)
@@ -579,8 +545,7 @@ class GAT:
                 )
             x_inp, predictions = gat.in_out_tensors()
 
-    For more details, please see the GAT demo notebook:
-    demos/node-classification/gat/gat-cora-node-classification-example.ipynb
+    For more details, please see `the GAT demo notebook <https://stellargraph.readthedocs.io/en/stable/demos/node-classification/gat-node-classification.html>`_
 
     Notes:
       - The inputs are tensors with a batch dimension of 1. These are provided by the \
@@ -756,7 +721,7 @@ class GAT:
         if not isinstance(activations, list):
             raise TypeError(
                 "{}: activations should be a list of strings; received {} instead".format(
-                    type(self).__name__, type(activations)
+                    type(self).__name__, type(activations).__name__
                 )
             )
         # check length:
@@ -829,7 +794,6 @@ class GAT:
                     attn_dropout_rate=self.attn_dropout,
                     activation=self.activations[ii],
                     use_bias=self.bias,
-                    final_layer=ii == (n_layers - 1),
                     saliency_map_support=self.saliency_map_support,
                     kernel_initializer=kernel_initializer,
                     kernel_regularizer=kernel_regularizer,
@@ -889,9 +853,8 @@ class GAT:
         h_layer = x_in
         for layer in self._layers:
             if isinstance(layer, self._gat_layer):
-                # For a GAT layer add the matrix and output indices
-                # Note that the output indices are only used if `final_layer=True`
-                h_layer = layer([h_layer, out_indices] + Ainput)
+                # For a GAT layer add the matrix
+                h_layer = layer([h_layer] + Ainput)
 
             else:
                 # For other (non-graph) layers only supply the input tensor
@@ -899,6 +862,9 @@ class GAT:
 
             # print("Hlayer:", h_layer)
 
+        # only return data for the requested nodes
+        h_layer = GatherIndices(batch_dims=1)([h_layer, out_indices])
+
         return self._normalization(h_layer)
 
     def in_out_tensors(self, multiplicity=None):
