@@ -14,6 +14,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import tensorflow as tf
 from tensorflow.keras.layers import Layer
 from tensorflow.keras import backend as K
 import warnings
@@ -96,6 +97,45 @@ class SqueezedSparseConversion(Layer):
         return output
 
 
+class GatherIndices(Layer):
+    """
+    Gathers slices from a data tensor, based on an indices tensors (``tf.gather`` in Layer form).
+
+    Args:
+        axis (int or Tensor): the data axis to gather from.
+        batch_dims (int): the number of batch dimensions in the data and indices.
+    """
+
+    def __init__(self, axis=None, batch_dims=0, **kwargs):
+        super().__init__(**kwargs)
+        self._axis = axis
+        self._batch_dims = batch_dims
+
+    def get_config(self):
+        config = super().get_config()
+        config.update(axis=self._axis, batch_dims=self._batch_dims)
+        return config
+
+    def compute_output_shape(self, input_shapes):
+        data_shape, indices_shape = input_shapes
+        axis = self._batch_dims if self._axis is None else self._axis
+        # per https://www.tensorflow.org/api_docs/python/tf/gather
+        return (
+            data_shape[:axis]
+            + indices_shape[self._batch_dims :]
+            + data_shape[axis + 1 :]
+        )
+
+    def call(self, inputs):
+        """
+        Args:
+            inputs (list): a pair of tensors, corresponding to the ``params`` and ``indices``
+                parameters to ``tf.gather``.
+        """
+        data, indices = inputs
+        return tf.gather(data, indices, axis=self._axis, batch_dims=self._batch_dims)
+
+
 def deprecated_model_function(function, old_name):
     def _function_wrapper(*args, **kwargs):
         """Deprecated: use :meth:`in_out_tensors`."""
