@@ -47,9 +47,10 @@ from ..data import (
 from ..core.utils import is_real_iterable
 from . import LinkSequence, OnDemandLinkSequence
 from ..random import SeededPerBatch
+from .base import Generator
 
 
-class BatchedLinkGenerator(abc.ABC):
+class BatchedLinkGenerator(Generator):
     def __init__(self, G, batch_size, schema=None):
         if not isinstance(G, StellarGraph):
             raise TypeError("Graph must be a StellarGraph or StellarDiGraph object.")
@@ -81,6 +82,9 @@ class BatchedLinkGenerator(abc.ABC):
     def sample_features(self, head_links, batch_num):
         pass
 
+    def num_batch_dims(self):
+        return 1
+
     def flow(self, link_ids, targets=None, shuffle=False, seed=None):
         """
         Creates a generator/sequence object for training or evaluation
@@ -102,7 +106,7 @@ class BatchedLinkGenerator(abc.ABC):
         Args:
             link_ids: an iterable of tuples of node IDs as (source, target)
             targets: a 2D array of numeric targets with shape
-                `(len(link_ids), target_size)`
+                ``(len(link_ids), target_size)``
             shuffle (bool): If True the links will be shuffled at each
                 epoch, if False the links will be processed in order.
             seed (int, optional): Random seed
@@ -150,6 +154,8 @@ class BatchedLinkGenerator(abc.ABC):
                         f"Node pair ({src}, {dst}) not of expected type ({expected_src_type}, {expected_dst_type})"
                     )
 
+            link_ids = [self.graph.node_ids_to_ilocs(ids) for ids in link_ids]
+
             return LinkSequence(
                 self.sample_features,
                 self.batch_size,
@@ -197,11 +203,9 @@ class GraphSAGELinkGenerator(BatchedLinkGenerator):
     At minimum, supply the StellarGraph, the batch size, and the number of
     node samples for each layer of the GraphSAGE model.
 
-    The supplied graph should be a StellarGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarGraph object with node features.
 
-    Use the :meth:`.flow` method supplying the nodes and (optionally) targets,
+    Use the :meth:`flow` method supplying the nodes and (optionally) targets,
     or an UnsupervisedSampler instance that generates node samples on demand,
     to get an object that can be used as a Keras data generator.
 
@@ -287,7 +291,7 @@ class GraphSAGELinkGenerator(BatchedLinkGenerator):
             # Get features for the sampled nodes
             batch_feats.append(
                 [
-                    self.graph.node_features(layer_nodes, node_type)
+                    self.graph.node_features(layer_nodes, node_type, use_ilocs=True,)
                     for layer_nodes in nodes_per_hop
                 ]
             )
@@ -310,9 +314,7 @@ class HinSAGELinkGenerator(BatchedLinkGenerator):
     At minimum, supply the StellarGraph, the batch size, and the number of
     node samples for each layer of the GraphSAGE model.
 
-    The supplied graph should be a StellarGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarGraph object with node features for all node types.
 
     Use the :meth:`flow` method supplying the nodes and (optionally) targets
     to get an object that can be used as a Keras data generator.
@@ -327,7 +329,8 @@ class HinSAGELinkGenerator(BatchedLinkGenerator):
         g (StellarGraph): A machine-learning ready graph.
         batch_size (int): Size of batch of links to return.
         num_samples (list): List of number of neighbour node samples per GraphSAGE layer (hop) to take.
-        head_node_types (list): List of the types (str) of the two head nodes forming the node pair.
+        head_node_types (list, optional): List of the types (str) of the two head nodes forming the
+            node pair. This does not need to be specified if ``G`` has only one node type.
         seed (int or str, optional): Random seed for the sampling methods.
 
     Example::
@@ -341,7 +344,7 @@ class HinSAGELinkGenerator(BatchedLinkGenerator):
         G,
         batch_size,
         num_samples,
-        head_node_types,
+        head_node_types=None,
         schema=None,
         seed=None,
         name=None,
@@ -351,6 +354,13 @@ class HinSAGELinkGenerator(BatchedLinkGenerator):
         self.name = name
 
         # This is a link generator and requires two nodes per query
+        if head_node_types is None:
+            # infer the head node types, if this is a homogeneous-node graph
+            node_type = G.unique_node_type(
+                "head_node_types: expected a pair of head node types because G has more than one node type, found node types: %(found)s"
+            )
+            head_node_types = [node_type, node_type]
+
         self.head_node_types = head_node_types
         if len(self.head_node_types) != 2:
             raise ValueError(
@@ -370,7 +380,7 @@ class HinSAGELinkGenerator(BatchedLinkGenerator):
             G, graph_schema=self.schema, seed=seed
         )
 
-    def _get_features(self, node_samples, head_size):
+    def _get_features(self, node_samples, head_size, use_ilocs=False):
         """
         Collect features from sampled nodes.
         Args:
@@ -385,7 +395,7 @@ class HinSAGELinkGenerator(BatchedLinkGenerator):
         # Resize features to (batch_size, n_neighbours, feature_size)
         # for each node type (note that we can have different feature size for each node type)
         batch_feats = [
-            self.graph.node_features(layer_nodes, nt)
+            self.graph.node_features(layer_nodes, nt, use_ilocs=use_ilocs)
             for nt, layer_nodes in node_samples
         ]
 
@@ -406,7 +416,7 @@ class HinSAGELinkGenerator(BatchedLinkGenerator):
 
         Returns:
             A list of the same length as `num_samples` of collected features from
-            the sampled nodes of shape: `(len(head_nodes), num_sampled_at_layer, feature_size)`
+            the sampled nodes of shape: ``(len(head_nodes), num_sampled_at_layer, feature_size)``
             where num_sampled_at_layer is the cumulative product of `num_samples`
             for that layer.
         """
@@ -443,7 +453,7 @@ class HinSAGELinkGenerator(BatchedLinkGenerator):
             for ab in zip(nodes_by_type[0], nodes_by_type[1])
         ]
 
-        batch_feats = self._get_features(nodes_by_type, len(head_links))
+        batch_feats = self._get_features(nodes_by_type, len(head_links), use_ilocs=True)
 
         return batch_feats
 
@@ -454,11 +464,9 @@ class Attri2VecLinkGenerator(BatchedLinkGenerator):
 
     At minimum, supply the StellarGraph and the batch size.
 
-    The supplied graph should be a StellarGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarGraph object with node features.
 
-    Use the :meth:`.flow` method supplying the nodes and targets,
+    Use the :meth:`flow` method supplying the nodes and targets,
     or an UnsupervisedSampler instance that generates node samples on demand,
     to get an object that can be used as a Keras data generator.
 
@@ -494,8 +502,8 @@ class Attri2VecLinkGenerator(BatchedLinkGenerator):
 
         target_ids = [head_link[0] for head_link in head_links]
         context_ids = [head_link[1] for head_link in head_links]
-        target_feats = self.graph.node_features(target_ids)
-        context_feats = self.graph._get_index_for_nodes(context_ids)
+        target_feats = self.graph.node_features(target_ids, use_ilocs=True)
+        context_feats = np.array(context_ids)
         batch_feats = [target_feats, np.array(context_feats)]
 
         return batch_feats
@@ -508,11 +516,9 @@ class DirectedGraphSAGELinkGenerator(BatchedLinkGenerator):
     At minimum, supply the StellarDiGraph, the batch size, and the number of
     node samples (separately for in-nodes and out-nodes) for each layer of the GraphSAGE model.
 
-    The supplied graph should be a StellarDiGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarDiGraph object with node features.
 
-    Use the :meth:`.flow` method supplying the nodes and (optionally) targets,
+    Use the :meth:`flow` method supplying the nodes and (optionally) targets,
     or an UnsupervisedSampler instance that generates node samples on demand,
     to get an object that can be used as a Keras data generator.
 
@@ -595,7 +601,9 @@ class DirectedGraphSAGELinkGenerator(BatchedLinkGenerator):
                 nodes_in_slot = [
                     element for sample in node_samples for element in sample[slot]
                 ]
-                features_for_slot = self.graph.node_features(nodes_in_slot, node_type)
+                features_for_slot = self.graph.node_features(
+                    nodes_in_slot, node_type, use_ilocs=True,
+                )
 
                 features[slot] = np.reshape(
                     features_for_slot, (len(hns), -1, features_for_slot.shape[1])
