@@ -37,6 +37,7 @@ import scipy.sparse as sps
 from tensorflow.keras import backend as K
 from functools import reduce
 from tensorflow.keras.utils import Sequence
+from collections import defaultdict
 
 from ..data import (
     SampledBreadthFirstWalk,
@@ -45,17 +46,15 @@ from ..data import (
 )
 from ..core.graph import StellarGraph, GraphSchema
 from ..core.utils import is_real_iterable
-from . import NodeSequence
+from . import NodeSequence, Generator
 from ..random import SeededPerBatch
 
 
-class BatchedNodeGenerator(abc.ABC):
+class BatchedNodeGenerator(Generator):
     """
     Abstract base class for graph data generators.
 
-    The supplied graph should be a StellarGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarGraph object with node features.
 
     Do not use this base class: use a subclass specific to the method.
 
@@ -96,6 +95,9 @@ class BatchedNodeGenerator(abc.ABC):
     def sample_features(self, head_nodes, batch_num):
         pass
 
+    def num_batch_dims(self):
+        return 1
+
     def flow(self, node_ids, targets=None, shuffle=False, seed=None):
         """
         Creates a generator/sequence object for training or evaluation
@@ -117,7 +119,7 @@ class BatchedNodeGenerator(abc.ABC):
         Args:
             node_ids: an iterable of node IDs
             targets: a 2D array of numeric targets with shape
-                `(len(node_ids), target_size)`
+                ``(len(node_ids), target_size)``
             shuffle (bool): If True the node_ids will be shuffled at each
                 epoch, if False the node_ids will be processed in order.
 
@@ -144,10 +146,11 @@ class BatchedNodeGenerator(abc.ABC):
                     f"Node ID {n} not of expected type {expected_node_type}"
                 )
 
+        node_ilocs = self.graph.node_ids_to_ilocs(node_ids)
         return NodeSequence(
             self.sample_features,
             self.batch_size,
-            node_ids,
+            node_ilocs,
             targets,
             shuffle=shuffle,
             seed=seed,
@@ -180,9 +183,7 @@ class GraphSAGENodeGenerator(BatchedNodeGenerator):
     At minimum, supply the StellarGraph, the batch size, and the number of
     node samples for each layer of the GraphSAGE model.
 
-    The supplied graph should be a StellarGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarGraph object with node features.
 
     Use the :meth:`flow` method supplying the nodes and (optionally) targets
     to get an object that can be used as a Keras data generator.
@@ -260,7 +261,7 @@ class GraphSAGENodeGenerator(BatchedNodeGenerator):
 
         # Get features for sampled nodes
         batch_feats = [
-            self.graph.node_features(layer_nodes, node_type)
+            self.graph.node_features(layer_nodes, node_type, use_ilocs=True)
             for layer_nodes in nodes_per_hop
         ]
 
@@ -271,6 +272,10 @@ class GraphSAGENodeGenerator(BatchedNodeGenerator):
         ]
         return batch_feats
 
+    def default_corrupt_input_index_groups(self):
+        # everything can be shuffled together
+        return [list(range(len(self.num_samples) + 1))]
+
 
 class DirectedGraphSAGENodeGenerator(BatchedNodeGenerator):
     """
@@ -281,9 +286,7 @@ class DirectedGraphSAGENodeGenerator(BatchedNodeGenerator):
     node samples (separately for in-nodes and out-nodes)
     for each layer of the GraphSAGE model.
 
-    The supplied graph should be a StellarDiGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarDiGraph object with node features.
 
     Use the :meth:`flow` method supplying the nodes and (optionally) targets
     to get an object that can be used as a Keras data generator.
@@ -324,6 +327,10 @@ class DirectedGraphSAGENodeGenerator(BatchedNodeGenerator):
             G, graph_schema=self.schema, seed=seed
         )
 
+    def _max_slots(self):
+        max_hops = len(self.in_samples)
+        return 2 ** (max_hops + 1) - 1
+
     def sample_features(self, head_nodes, batch_num):
         """
         Sample neighbours recursively from the head nodes, collect the features of the
@@ -352,13 +359,14 @@ class DirectedGraphSAGENodeGenerator(BatchedNodeGenerator):
 
         node_type = self.head_node_types[0]
 
-        max_hops = len(self.in_samples)
-        max_slots = 2 ** (max_hops + 1) - 1
+        max_slots = self._max_slots()
         features = [None] * max_slots  # flattened binary tree
 
         for slot in range(max_slots):
             nodes_in_slot = list(it.chain(*[sample[slot] for sample in node_samples]))
-            features_for_slot = self.graph.node_features(nodes_in_slot, node_type)
+            features_for_slot = self.graph.node_features(
+                nodes_in_slot, node_type, use_ilocs=True
+            )
             resize = -1 if np.size(features_for_slot) > 0 else 0
             features[slot] = np.reshape(
                 features_for_slot, (len(head_nodes), resize, features_for_slot.shape[1])
@@ -366,6 +374,10 @@ class DirectedGraphSAGENodeGenerator(BatchedNodeGenerator):
 
         return features
 
+    def default_corrupt_input_index_groups(self):
+        # everything can be shuffled together
+        return [list(range(self._max_slots()))]
+
 
 class HinSAGENodeGenerator(BatchedNodeGenerator):
     """Keras-compatible data mapper for Heterogeneous GraphSAGE (HinSAGE)
@@ -373,9 +385,7 @@ class HinSAGENodeGenerator(BatchedNodeGenerator):
     At minimum, supply the StellarGraph, the batch size, and the number of
     node samples for each layer of the HinSAGE model.
 
-    The supplied graph should be a StellarGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarGraph object with node features for all node types.
 
     Use the :meth:`flow` method supplying the nodes and (optionally) targets
     to get an object that can be used as a Keras data generator.
@@ -387,8 +397,9 @@ class HinSAGENodeGenerator(BatchedNodeGenerator):
         G (StellarGraph): The machine-learning ready graph
         batch_size (int): Size of batch to return
         num_samples (list): The number of samples per layer (hop) to take
-        head_node_type (str): The node type that will be given to the generator
-            using the `flow` method, the model will expect this node type.
+        head_node_type (str, optional): The node type that will be given to the generator using the
+            `flow` method, the model will expect this node type. This does not need to be specified
+            if ``G`` has only one node type.
         schema (GraphSchema, optional): Graph schema for G.
         seed (int, optional): Random seed for the node sampler
 
@@ -405,7 +416,7 @@ class HinSAGENodeGenerator(BatchedNodeGenerator):
         G,
         batch_size,
         num_samples,
-        head_node_type,
+        head_node_type=None,
         schema=None,
         seed=None,
         name=None,
@@ -416,6 +427,12 @@ class HinSAGENodeGenerator(BatchedNodeGenerator):
         self.name = name
 
         # The head node type
+        if head_node_type is None:
+            # infer the head node type, if this is a homogeneous-node graph
+            head_node_type = G.unique_node_type(
+                "head_node_type: expected a head node type because G has more than one node type, found node types: %(found)s"
+            )
+
         if head_node_type not in self.schema.node_types:
             raise KeyError("Supplied head node type must exist in the graph")
         self.head_node_types = [head_node_type]
@@ -469,7 +486,7 @@ class HinSAGENodeGenerator(BatchedNodeGenerator):
 
         # Get features
         batch_feats = [
-            self.graph.node_features(layer_nodes, nt)
+            self.graph.node_features(layer_nodes, nt, use_ilocs=True)
             for nt, layer_nodes in nodes_by_type
         ]
 
@@ -481,6 +498,16 @@ class HinSAGENodeGenerator(BatchedNodeGenerator):
 
         return batch_feats
 
+    def default_corrupt_input_index_groups(self):
+        # every sample of a given node type can be grouped together
+        indices_per_nt = defaultdict(list)
+        for tensor_idx, (nt, _) in enumerate(self._sampling_schema[0]):
+            indices_per_nt[nt].append(tensor_idx)
+
+        # ensure there's a consistent order both within each group, and across groups, ensure the
+        # shuffling is deterministic (at least with respect to the model)
+        return sorted(sorted(idx) for idx in indices_per_nt.values())
+
 
 class Attri2VecNodeGenerator(BatchedNodeGenerator):
     """
@@ -489,9 +516,7 @@ class Attri2VecNodeGenerator(BatchedNodeGenerator):
 
     At minimum, supply the StellarGraph and the batch size.
 
-    The supplied graph should be a StellarGraph object that is ready for
-    machine learning. Currently the model requires node features for all
-    nodes in the graph.
+    The supplied graph should be a StellarGraph object with node features.
 
     Use the :meth:`flow` method supplying the nodes to get an object
     that can be used as a Keras data generator.
@@ -525,7 +550,7 @@ class Attri2VecNodeGenerator(BatchedNodeGenerator):
             head node.
         """
 
-        batch_feats = self.graph.node_features(head_nodes)
+        batch_feats = self.graph.node_features(head_nodes, use_ilocs=True)
         return batch_feats
 
     def flow(self, node_ids):
@@ -545,8 +570,9 @@ class Attri2VecNodeGenerator(BatchedNodeGenerator):
             in the Keras method ``predict``.
 
         """
+        node_ilocs = self.graph.node_ids_to_ilocs(node_ids)
         return NodeSequence(
-            self.sample_features, self.batch_size, node_ids, shuffle=False
+            self.sample_features, self.batch_size, node_ilocs, shuffle=False
         )
 
     def flow_from_dataframe(self, node_ids):
