@@ -30,9 +30,9 @@ _EMPTY_DF = pd.DataFrame([], index=[1, 2])
 
 
 def test_columnar_convert_type_default():
-    converter = ColumnarConverter("some_name", "foo", {}, {}, False)
-    shared, features = converter.convert(_EMPTY_DF)
-    assert "foo" in shared
+    converter = ColumnarConverter("some_name", "foo", None, {}, {}, False, {})
+    shared, type_starts, features = converter.convert(_EMPTY_DF)
+    assert type_starts == [("foo", 0)]
     assert "foo" in features
 
 
@@ -40,22 +40,20 @@ def test_columnar_convert_selected_columns():
     df = _EMPTY_DF.assign(before="abc", same=10)
 
     converter = ColumnarConverter(
-        "some_name", "foo", {}, {"before": "after", "same": "same"}, False
+        "some_name", "foo", None, {}, {"before": "after", "same": "same"}, False, {}
     )
-    shared, features = converter.convert({"x": df, "y": df})
+    shared, type_starts, features = converter.convert({"x": df, "y": df})
 
-    assert "x" in shared
-    assert "y" in shared
+    assert type_starts == [("x", 0), ("y", 2)]
 
-    for df in shared.values():
-        assert "before" not in df
-        assert all(df["after"] == "abc")
-        assert all(df["same"] == 10)
+    assert "before" not in shared
+    assert all(shared["after"] == "abc")
+    assert all(shared["same"] == 10)
 
 
 def test_columnar_convert_selected_columns_missing():
     converter = ColumnarConverter(
-        "some_name", "foo", {}, {"before": "after", "same": "same"}, False
+        "some_name", "foo", None, {}, {"before": "after", "same": "same"}, False, {}
     )
 
     with pytest.raises(
@@ -65,62 +63,199 @@ def test_columnar_convert_selected_columns_missing():
 
 
 def test_columnar_convert_column_default():
-    converter = ColumnarConverter("some_name", "foo", {"before": 123}, {}, False)
-    shared, features = converter.convert({"x": _EMPTY_DF, "y": _EMPTY_DF})
+    converter = ColumnarConverter(
+        "some_name", "foo", None, {"before": 123}, {}, False, {}
+    )
+    shared, type_starts, features = converter.convert({"x": _EMPTY_DF, "y": _EMPTY_DF})
 
-    assert "x" in shared
-    assert "y" in shared
+    assert type_starts == [("x", 0), ("y", 2)]
 
-    for df in shared.values():
-        assert all(df["before"] == 123)
+    assert all(shared["before"] == 123)
 
 
 def test_columnar_convert_column_default_selected_columns():
     # the defaulting happens before the renaming
     converter = ColumnarConverter(
-        "x", "foo", {"before": 123}, {"before": "after"}, False
+        "x", "foo", None, {"before": 123}, {"before": "after"}, False, {}
     )
-    shared, features = converter.convert({"x": _EMPTY_DF, "y": _EMPTY_DF})
+    shared, type_starts, features = converter.convert({"x": _EMPTY_DF, "y": _EMPTY_DF})
 
-    assert "x" in shared
-    assert "y" in shared
+    assert type_starts == [("x", 0), ("y", 2)]
 
-    for df in shared.values():
-        assert "before" not in df
-        assert all(df["after"] == 123)
+    assert "before" not in shared
+    assert all(shared["after"] == 123)
 
 
 def test_columnar_convert_features():
-    converter = ColumnarConverter("some_name", "foo", {}, {"x": "x"}, True)
+    converter = ColumnarConverter("some_name", "foo", None, {}, {"x": "x"}, True, {})
     df = _EMPTY_DF.assign(a=[1, 2], b=[100, 200], x=123)
-    shared, features = converter.convert(df)
+    shared, type_starts, features = converter.convert(df)
 
-    assert all(shared["foo"]["x"] == 123)
+    assert type_starts == [("foo", 0)]
+    assert all(shared["x"] == 123)
     assert np.array_equal(features["foo"], [[1, 100], [2, 200]])
 
 
 def test_columnar_convert_disallow_features():
-    converter = ColumnarConverter("some_name", "foo", {}, {}, False)
+    converter = ColumnarConverter("some_name", "foo", None, {}, {}, False, {})
     df = _EMPTY_DF.assign(a=1)
     with pytest.raises(ValueError, match="expected zero feature columns, found 'a'"):
-        shared, features = converter.convert(df)
+        shared, type_starts, features = converter.convert(df)
 
 
 def test_columnar_convert_invalid_input():
-    converter = ColumnarConverter("some_name", "foo", {}, {}, False)
+    converter = ColumnarConverter("some_name", "foo", None, {}, {}, False, {})
 
-    with pytest.raises(
-        TypeError, match="some_name: expected dict, found <class 'int'>"
-    ):
+    with pytest.raises(TypeError, match="some_name: expected dict, found int"):
         converter.convert(1)
 
     with pytest.raises(
-        TypeError,
-        match=r"some_name\['x'\]: expected pandas DataFrame, found <class 'int'>",
+        TypeError, match=r"some_name\['x'\]: expected pandas DataFrame, found int",
     ):
         converter.convert({"x": 1})
 
 
+def test_columnar_convert_type_column():
+    converter = ColumnarConverter(
+        "some_name",
+        "foo",
+        "type_column",
+        {},
+        {"type_column": "TC", "data": "D"},
+        False,
+        {},
+    )
+
+    df = pd.DataFrame(
+        {"type_column": ["c", "a", "a", "c", "b"], "data": [1, 2, 3, 4, 5]},
+        index=[1, 10, 100, 1000, 10000],
+    )
+    shared, type_starts, features = converter.convert(df)
+
+    assert set(shared.columns) == {"D"}
+    assert list(shared.index) == [10, 100, 10000, 1, 1000]
+    assert list(shared["D"]) == [2, 3, 5, 1, 4]
+    assert type_starts == [("a", 0), ("b", 2), ("c", 3)]
+    assert features == {"a": None, "b": None, "c": None}
+
+    # invalid configurations
+    with pytest.raises(
+        ValueError, match=r"allow_features: expected no features .* \('type_column'\)"
+    ):
+        ColumnarConverter(
+            "some_name", "foo", "type_column", {}, {"type_column": "TC"}, True, {}
+        )
+
+    with pytest.raises(
+        ValueError,
+        match=r"selected_columns: expected type column \('type_column'\) .* found only 'TC', 'data'",
+    ):
+        ColumnarConverter(
+            "some_name",
+            "foo",
+            "type_column",
+            {},
+            {"TC": "type_column", "data": "D"},
+            False,
+            {},
+        )
+
+
+def test_columnar_convert_transform_columns():
+
+    columns = {"x": np.complex128(1), "y": np.uint16(2), "z": np.float32(3.0)}
+
+    dfs = {
+        name: pd.DataFrame({"s": [0], "t": [1], "w": [w]}, index=[i])
+        for i, (name, w) in enumerate(columns.items())
+    }
+
+    converter = ColumnarConverter(
+        "some_name",
+        float,
+        None,
+        column_defaults={},
+        selected_columns={"s": "ss", "t": "tt", "w": "ww",},
+        transform_columns={"w": lambda x: x + 1,},
+        allow_features=False,
+    )
+
+    converted, type_starts, _ = converter.convert(dfs)
+
+    assert (converted.iloc[type_starts[0][1] : type_starts[1][1]]["ww"] == 2).all()
+    assert (converted[type_starts[1][1] : type_starts[2][1]]["ww"] == 3).all()
+    assert (converted[type_starts[2][1] :]["ww"] == 4).all()
+
+    assert (converted["ss"] == 0).all()
+    assert (converted["tt"] == 1).all()
+
+
+def test_convert_edges_weights():
+    def run(ws):
+        dfs = {
+            name: pd.DataFrame({"s": [0], "t": [1], "w": [w]}, index=[i])
+            for i, (name, w) in enumerate(ws.items())
+        }
+        nodes = convert_nodes(
+            pd.DataFrame([], index=[0, 1]),
+            name="other_name",
+            default_type=np.int8,
+            dtype=np.int8,
+        )
+
+        convert_edges(
+            dfs,
+            name="some_name",
+            default_type="d",
+            source_column="s",
+            target_column="t",
+            weight_column="w",
+            nodes=nodes,
+            type_column=None,
+        )
+
+    # various numbers are valid
+    run({"x": np.int8(1)})
+    run({"x": np.complex64(1)})
+    run({"x": np.complex128(1), "y": np.uint16(2), "z": np.float32(3.0)})
+    # non-numbers are not
+    with pytest.raises(
+        TypeError,
+        match=r"some_name: expected weight column 'w' to be numeric, found dtype 'object'",
+    ):
+        run({"x": "ab", "y": 1, "z": np.float32(2)})
+
+
+def test_convert_edges_type_column():
+    data = pd.DataFrame(
+        {
+            "s": [10, 20, 30, 40, 50],
+            "t": [20, 30, 40, 50, 60],
+            "l": ["c", "a", "a", "c", "b"],
+        }
+    )
+
+    nodes = pd.DataFrame([], index=[10, 20, 30, 40, 50, 60])
+    nodes = convert_nodes(nodes, name="other_name", default_type=np.int8, dtype=np.int8)
+
+    edges = convert_edges(
+        data,
+        name="some_name",
+        default_type="d",
+        source_column="s",
+        target_column="t",
+        weight_column="w",
+        type_column="l",
+        nodes=nodes,
+    )
+
+    np.testing.assert_array_equal(edges.sources, [1, 2, 4, 0, 3])
+    np.testing.assert_array_equal(edges.targets, [2, 3, 5, 1, 4])
+    np.testing.assert_array_equal(
+        edges.type_of_iloc(slice(None)), ["a", "a", "b", "c", "c"]
+    )
+
+
 def from_networkx_for_testing(g, node_features=None, dtype="float32"):
     return from_networkx(
         g,
@@ -356,6 +491,29 @@ def test_from_networkx_default_features():
     assert edges == {}
 
 
+def test_from_networkx_empty():
+    g = nx.Graph()
+    # no features, means no types
+    nodes, edges = from_networkx_for_testing(g)
+    assert nodes == {}
+    assert edges == {}
+
+    # various forms of features:
+    nodes, edges = from_networkx_for_testing(g, node_features={})
+    assert nodes == {}
+    assert edges == {}
+
+    features = pd.DataFrame(columns=range(10))
+
+    nodes, edges = from_networkx_for_testing(g, node_features=features)
+    assert nodes == {}
+    assert edges == {}
+
+    nodes, edges = from_networkx_for_testing(g, node_features={"b": features})
+    assert nodes == {}
+    assert edges == {}
+
+
 def test_from_networkx_errors():
     het = nx.MultiDiGraph()
     het.add_nodes_from([(0, {"n": "a"}), (1, {"n": "b"})])
