@@ -20,7 +20,7 @@ import pytest
 import networkx as nx
 from stellargraph.data.explorer import BiasedRandomWalk
 from stellargraph.core.graph import StellarGraph
-from ..test_utils.graphs import create_test_graph
+from ..test_utils.graphs import create_test_graph, example_graph_random
 
 
 # FIXME (#535): Consider using graph fixtures
@@ -90,6 +90,24 @@ class TestBiasedWeightedRandomWalk(object):
                 nodes=nodes, n=n, p=p, q=q, length=length, seed=seed, weighted="unknown"
             )
 
+    def test_init_parameters(self):
+        g = weighted(1, 2, 3, 4)
+        nodes = list(g.nodes())
+        n = 4
+        length = 4
+        seed = 42
+        p = 1.0
+        q = 1.0
+
+        rw = BiasedRandomWalk(g, n=n, p=p, q=q, length=length, seed=seed, weighted=True)
+        rw_no_params = BiasedRandomWalk(g)
+
+        run_1 = rw.run(nodes=nodes)
+        run_2 = rw_no_params.run(
+            nodes=nodes, n=n, p=p, q=q, length=length, seed=seed, weighted=True
+        )
+        assert np.array_equal(run_1, run_2)
+
     def test_identity_unweighted_weighted_1_walks(self):
 
         # graph with all edge weights = 1
@@ -102,11 +120,13 @@ class TestBiasedWeightedRandomWalk(object):
         q = 1.0
 
         biasedrw = BiasedRandomWalk(g)
-        assert biasedrw.run(
+        run_1 = biasedrw.run(
             nodes=nodes, n=n, p=p, q=q, length=length, seed=seed, weighted=True
-        ) == biasedrw.run(
-            nodes=nodes, n=n, p=p, q=q, length=length, seed=seed, weighted=False
         )
+        run_2 = biasedrw.run(
+            nodes=nodes, n=n, p=p, q=q, length=length, seed=seed, weighted=True
+        )
+        assert np.array_equal(run_1, run_2)
 
     def test_weighted_walks(self):
 
@@ -135,7 +155,12 @@ class TestBiasedWeightedRandomWalk(object):
 
         biasedrw = BiasedRandomWalk(g)
 
-        with pytest.raises(ValueError):
+        neg_message = r"graph: expected all edge weights to be non-negative and finite, found some negative or infinite: 1 to 2 \(weight = -2\)"
+
+        with pytest.raises(ValueError, match=neg_message):
+            BiasedRandomWalk(g, weighted=True)
+
+        with pytest.raises(ValueError, match=neg_message):
             biasedrw.run(
                 nodes=nodes, n=n, p=p, q=q, length=length, seed=seed, weighted=True
             )
@@ -145,7 +170,11 @@ class TestBiasedWeightedRandomWalk(object):
 
         biasedrw = BiasedRandomWalk(g)
 
-        with pytest.raises(ValueError):
+        inf_message = r"graph: expected all edge weights to be non-negative and finite, found some negative or infinite: 1 to 2 \(weight = inf\)"
+        with pytest.raises(ValueError, match=inf_message):
+            BiasedRandomWalk(g, weighted=True)
+
+        with pytest.raises(ValueError, match=inf_message):
             biasedrw.run(
                 nodes=nodes, n=n, p=p, q=q, length=length, seed=seed, weighted=True
             )
@@ -189,11 +218,11 @@ class TestBiasedWeightedRandomWalk(object):
         g[4][1]["wt"] = 4
 
     def test_benchmark_biasedweightedrandomwalk(self, benchmark):
-        g = create_test_weighted_graph()
+        g = example_graph_random(n_nodes=100, n_edges=500)
         biasedrw = BiasedRandomWalk(g)
 
-        nodes = ["0"]
-        n = 5
+        nodes = np.arange(0, 50)
+        n = 2
         p = 2
         q = 3
         length = 5
@@ -230,13 +259,13 @@ class TestBiasedRandomWalk(object):
             biasedrw.run(nodes=nodes, n=0, p=p, q=q, length=length, seed=seed)
         with pytest.raises(ValueError):
             biasedrw.run(nodes=nodes, n=-121, p=p, q=q, length=length, seed=seed)
-        with pytest.raises(ValueError):
+        with pytest.raises(TypeError):
             biasedrw.run(nodes=nodes, n=21.4, p=p, q=q, length=length, seed=seed)
-        with pytest.raises(ValueError):
+        with pytest.raises(TypeError):
             biasedrw.run(nodes=nodes, n=-0.5, p=p, q=q, length=length, seed=seed)
-        with pytest.raises(ValueError):
+        with pytest.raises(TypeError):
             biasedrw.run(nodes=nodes, n=0.0001, p=p, q=q, length=length, seed=seed)
-        with pytest.raises(ValueError):
+        with pytest.raises(TypeError):
             biasedrw.run(nodes=nodes, n="2", p=p, q=q, length=length, seed=seed)
 
         # p has to be > 0.
@@ -260,17 +289,17 @@ class TestBiasedRandomWalk(object):
             biasedrw.run(nodes=nodes, n=n, p=p, q=q, length=0, seed=seed)
         with pytest.raises(ValueError):
             biasedrw.run(nodes=nodes, n=n, p=p, q=q, length=-5, seed=seed)
-        with pytest.raises(ValueError):
+        with pytest.raises(TypeError):
             biasedrw.run(nodes=nodes, n=n, p=p, q=q, length=11.9, seed=seed)
-        with pytest.raises(ValueError):
+        with pytest.raises(TypeError):
             biasedrw.run(nodes=nodes, n=n, p=p, q=q, length=-9.9, seed=seed)
-        with pytest.raises(ValueError):
+        with pytest.raises(TypeError):
             biasedrw.run(nodes=nodes, n=n, p=p, q=q, length="10", seed=seed)
 
         # seed has to be None, 0,  or positive integer
         with pytest.raises(ValueError):
             biasedrw.run(nodes=nodes, n=n, p=p, q=q, length=length, seed=-1)
-        with pytest.raises(ValueError):
+        with pytest.raises(TypeError):
             biasedrw.run(nodes=nodes, n=n, p=p, q=q, length=length, seed=1010.8)
 
         # If no root nodes are given, an empty list is returned which is not an error but I thought this method
@@ -507,11 +536,11 @@ class TestBiasedRandomWalk(object):
         }
 
     def test_benchmark_biasedrandomwalk(self, benchmark):
-        g = create_test_graph()
+        g = example_graph_random(n_nodes=100, n_edges=500)
         biasedrw = BiasedRandomWalk(g)
 
-        nodes = ["0"]
-        n = 5
+        nodes = np.arange(0, 50)
+        n = 2
         p = 2
         q = 3
         length = 5
