@@ -220,6 +220,46 @@ def test_APPNP_apply_propagate_model_dense():
     assert preds_1 == pytest.approx(preds_2)
 
 
+@pytest.mark.parametrize("model_type", ["sequential", "model"])
+def test_APPNP_propagate_model_matches_manual(model_type):
+    dense_size = 5
+
+    G, features = create_graph_features()
+    adj = G.to_adjacency_matrix()
+    features, adj = GCN_Aadj_feats_op(features, adj)
+    adj = np.array(adj.todense()[None, :, :])
+    out_indices = np.array([[0, 1]], dtype="int32")
+
+    generator = FullBatchNodeGenerator(G, sparse=False, method="gcn")
+    appnpnModel = APPNP(
+        [dense_size], generator=generator, activations=["relu"], dropout=0.0
+    )
+
+    dense = Dense(dense_size)
+    if model_type == "sequential":
+        fully_connected_model = keras.Sequential()
+        fully_connected_model.add(dense)
+    else:
+        inp = keras.Input(shape=features.shape)
+        fully_connected_model = keras.Model(inp, dense(inp))
+
+    x_in, x_out = appnpnModel.propagate_model(fully_connected_model)
+    end_to_end_model = keras.Model(inputs=x_in, outputs=x_out)
+    preds_1 = end_to_end_model.predict([features[None, :, :], out_indices, adj])
+
+    # run the process manually: transform the features, and then propagate
+    float_feats = features[None, :, :].astype("float32")
+    manual_preds = manual_inp = fully_connected_model.predict(float_feats)
+    propagate = APPNPPropagationLayer(dense_size, teleport_probability=0.1)
+    for _ in range(10):
+        manual_preds = propagate([manual_preds, manual_inp, adj[0, ...]])
+
+    # select the relevant pieces
+    manual_preds = manual_preds.numpy()[:, out_indices.ravel(), :]
+
+    np.testing.assert_allclose(preds_1, manual_preds)
+
+
 def test_APPNP_apply_propagate_model_sparse():
 
     G, features = create_graph_features()
