@@ -589,6 +589,68 @@ def test_hinnodemapper_no_neighbors():
     assert np.all(batch_feats[3][:, 0, 0] == np.array([12, 0, 0]))
 
 
+def test_hinsage_corrupt_indices():
+    # prime and different feature sizes, so we can be more sure that things are lining up correctly.
+    feature_sizes = {"t1": 7, "t2": 11}
+    G, _, nodes_type_2 = example_hin_3(feature_sizes)
+
+    mapper = HinSAGENodeGenerator(
+        G, batch_size=2, num_samples=[3, 5], head_node_type="t2"
+    )
+
+    seq = mapper.flow(nodes_type_2)
+    tensors, _targets = seq[0]
+
+    groups = mapper.default_corrupt_input_index_groups()
+
+    # one group per type
+    assert len(groups) == 2
+    # every input tensor is included
+    assert {idx for g in groups for idx in g} == set(range(len(tensors)))
+
+    # check each group corresponds to nodes of a single type (i.e. the feature dimension for the
+    # tensors in each group are all the same), t2 is the head node so it should be first.
+    assert {tensors[idx].shape[-1] for idx in groups[0]} == {11}
+    assert {tensors[idx].shape[-1] for idx in groups[1]} == {7}
+
+
+def test_hinsage_homogeneous_inference():
+    feature_size = 4
+    edge_types = 3
+    batch_size = 2
+    num_samples = [5, 7]
+    G = example_graph_random(
+        feature_size=feature_size, node_types=1, edge_types=edge_types
+    )
+
+    # G is homogeneous so the head_node_type argument isn't required
+    mapper = HinSAGENodeGenerator(G, batch_size=batch_size, num_samples=num_samples)
+
+    assert mapper.head_node_types == ["n-0"]
+
+    nodes = [1, 4, 2]
+    seq = mapper.flow(nodes)
+    assert len(seq) == 2
+
+    samples_per_head = 1 + edge_types + edge_types * edge_types
+    for batch_idx, (samples, labels) in enumerate(seq):
+        this_batch_size = {0: batch_size, 1: 1}[batch_idx]
+
+        assert len(samples) == samples_per_head
+
+        assert samples[0].shape == (this_batch_size, 1, feature_size)
+        for i in range(1, 1 + edge_types):
+            assert samples[i].shape == (this_batch_size, num_samples[0], feature_size)
+        for i in range(1 + edge_types, samples_per_head):
+            assert samples[i].shape == (
+                this_batch_size,
+                np.product(num_samples),
+                feature_size,
+            )
+
+        assert labels is None
+
+
 def test_attri2vec_nodemapper_constructor_nx():
     """
     Attri2VecNodeGenerator requires a StellarGraph object
@@ -654,8 +716,8 @@ def test_attri2vec_nodemapper_1():
     assert nf.shape == (1, n_feat)
 
     # This will fail as the nodes are not in the graph
-    #    with pytest.raises(KeyError):
-    Attri2VecNodeGenerator(G1, batch_size=2).flow(["A", "B"])
+    with pytest.raises(KeyError):
+        Attri2VecNodeGenerator(G1, batch_size=2).flow(["A", "B"])
 
 
 def test_attri2vec_nodemapper_2():
@@ -697,7 +759,10 @@ class Test_FullBatchNodeGenerator:
     def test_generator_constructor_hin(self):
         feature_sizes = {"t1": 1, "t2": 1}
         Ghin, nodes_type_1, nodes_type_2 = example_hin_3(feature_sizes)
-        with pytest.raises(TypeError):
+        with pytest.raises(
+            ValueError,
+            match="G: expected a graph with a single node type, found a graph with node types: 't1', 't2'",
+        ):
             generator = FullBatchNodeGenerator(Ghin)
 
     def generator_flow(
@@ -902,30 +967,3 @@ class Test_FullBatchNodeGenerator:
             ppnp_sparse_failed = True
 
         assert ppnp_sparse_failed
-
-
-@pytest.mark.parametrize("sparse", [True, False])
-def test_corrupt_full_batch_generator(sparse):
-
-    G = example_graph_random(n_nodes=20)
-
-    generator = FullBatchNodeGenerator(G, sparse=sparse)
-
-    base_gen = generator.flow(G.nodes())
-    gen = CorruptedNodeSequence(base_gen)
-
-    [shuffled_feats, features, *_], targets = gen[0]
-
-    assert features.shape == shuffled_feats.shape
-
-    # check shuffled_feats are feats
-    assert not np.array_equal(features, shuffled_feats)
-
-    # check that all feature vecs in shuffled_feats correspond to a feature vec in features
-    assert all(
-        any(
-            np.array_equal(shuffled_feats[:, i, :], features[:, j, :])
-            for j in range(features.shape[1])
-        )
-        for i in range(shuffled_feats.shape[1])
-    )
